from typing import Any, List, Sequence, Tuple

import numpy as np
from numpy.typing import NDArray
from pyglm.glm import mat3, mat4, normalize, vec3


def create_sphere(
    radius: float = 1.0, rings: int = 16, sectors: int = 32
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    inv_r = 1.0 / (rings - 1)
    inv_s = 1.0 / (sectors - 1)

    r = np.arange(rings, dtype=np.float32)
    s = np.arange(sectors, dtype=np.float32)

    sin_r = np.sin(np.pi * r * inv_r)
    y_r = np.sin(-np.pi / 2 + np.pi * r * inv_r)
    cos_s = np.cos(2 * np.pi * s * inv_s)
    sin_s = np.sin(2 * np.pi * s * inv_s)

    # Broadcasting (rings, 1) * (1, sectors)
    x = (sin_r[:, None] * cos_s[None, :]).reshape(-1)
    y = np.repeat(y_r, sectors)
    z = (sin_r[:, None] * sin_s[None, :]).reshape(-1)

    normals = np.empty((rings * sectors, 3), dtype=np.float32)
    normals[:, 0] = x
    normals[:, 1] = y
    normals[:, 2] = z

    vertices = normals * radius

    r = np.arange(rings - 1, dtype=np.uint32)
    s = np.arange(sectors - 1, dtype=np.uint32)

    base = (r[:, None] * sectors + s[None, :]).reshape(-1)

    v0 = base
    v1 = base + sectors + 1
    v2 = base + 1
    v3 = base + sectors

    faces = np.empty(((rings - 1) * (sectors - 1) * 2, 3), dtype=np.uint32)
    faces[0::2, 0] = v0
    faces[0::2, 1] = v1
    faces[0::2, 2] = v2
    faces[1::2, 0] = v0
    faces[1::2, 1] = v3
    faces[1::2, 2] = v1

    return vertices, normals, faces.flatten()


def create_disk(
    radius: float = 1.0, sectors: int = 8
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    angle = 2 * np.pi / sectors

    vertices = np.zeros((sectors + 1, 3), np.float32)
    vertices[1:, 0] = radius * np.cos(np.arange(sectors, dtype=np.float32) * angle)
    vertices[1:, 1] = radius * np.sin(np.arange(sectors, dtype=np.float32) * angle)

    normals = np.zeros_like(vertices)
    normals[:, 2] = 1.0

    faces = np.zeros((sectors, 3), dtype=np.uint32)
    idxs = np.array(range(1, sectors + 1), dtype=np.uint32)
    faces[:, 1] = idxs
    faces[:-1, 2] = idxs[1:]
    faces[-1, 2] = 1

    return vertices, normals, faces.flatten()


def create_cylinder(
    radius: float = 1.0, height: float = 1.0, sectors: int = 8
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    angle = 2 * np.pi / sectors

    vertices = np.zeros((sectors * 2, 3), np.float32)
    vertices[:sectors, 0] = radius * np.cos(np.arange(sectors, dtype=np.float32) * angle)
    vertices[:sectors, 1] = radius * np.sin(np.arange(sectors, dtype=np.float32) * angle)
    vertices[sectors:, :2] = vertices[:sectors, :2]
    vertices[sectors:, 2] = height

    normals = np.zeros_like(vertices)
    normals[:, :2] = vertices[:, :2]
    normals = normals / np.linalg.norm(normals, axis=1)[..., np.newaxis]

    idxs_bot = np.arange(0, sectors, dtype=np.uint32)
    idxs_top = idxs_bot + sectors
    faces_top = np.zeros((sectors, 3), dtype=np.uint32)
    faces_top[:, 0] = idxs_top
    faces_top[:, 1] = idxs_bot
    faces_top[:-1, 2] = idxs_top[1:]
    faces_top[-1, 2] = idxs_top[0]
    faces_bot = np.zeros((sectors, 3), dtype=np.uint32)
    faces_bot[:, 0] = faces_top[:, 2]
    faces_bot[:, 1] = idxs_bot
    faces_bot[:-1, 2] = idxs_bot[1:]
    faces_bot[-1, 2] = idxs_bot[0]
    faces = np.concatenate([faces_top, faces_bot], axis=0)

    return vertices, normals, faces.flatten()


def create_capped_cylinder(
    radius: float = 1.0, height: float = 1.0, sectors: int = 8
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    cylinder_v, cylinder_n, cylinder_f = create_cylinder(radius, height, sectors)
    cap_v, cap_n, cap_f = create_disk(radius, sectors)

    # Translate top cap
    top_cap_v = cap_v + np.array([0.0, 0.0, height], np.float32)

    # Flip bot cap normals and windng order
    bot_cap_n = -cap_n
    bot_cap_f = cap_f.copy().reshape((-1, 3))
    bot_cap_f[:, (2, 1)] = bot_cap_f[:, (1, 2)]
    bot_cap_f = bot_cap_f.flatten()  # type: ignore

    (v, n), f = concatenate_meshes(
        [
            (cap_v, bot_cap_n),
            (top_cap_v, cap_n),
            (cylinder_v, cylinder_n),
        ],
        [bot_cap_f, cap_f, cylinder_f],
    )
    return v, n, f


def create_cone(
    radius: float = 1.0, height: float = 1.0, sectors: int = 8
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    angle = 2 * np.pi / sectors

    c = np.cos(np.arange(sectors, dtype=np.float32) * angle)
    s = np.sin(np.arange(sectors, dtype=np.float32) * angle)
    vertices = np.zeros((sectors + 1, 3), np.float32)
    vertices[:sectors, 0] = radius * c
    vertices[:sectors, 1] = radius * s
    vertices[sectors, 2] = height

    normal = np.array(normalize(vec3(height, 0.0, radius)))
    normals = np.zeros_like(vertices)
    normals[:sectors, 0] = c * normal[0]
    normals[:sectors, 1] = s * normal[0]

    faces = np.zeros((sectors, 3), dtype=np.uint32)
    faces[:, 0] = np.arange(sectors, dtype=np.uint32)
    faces[:-1, 1] = faces[1:, 0]
    faces[-1, 1] = 0
    faces[:, 2] = sectors
    return vertices, normals, faces.flatten()


def create_capped_cone(
    radius: float = 1.0, height: float = 1.0, sectors: int = 8
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    cap_v, cap_n, cap_f = create_disk(radius, sectors)
    cap_n = -cap_n
    cap_f = cap_f.copy().reshape((-1, 3))
    cap_f[:, (2, 1)] = cap_f[:, (1, 2)]
    cap_f = cap_f.flatten()

    cone_v, cone_n, cone_f = create_cone(radius, height, sectors)
    (v, n), f = concatenate_meshes(
        [
            (cap_v, cap_n),
            (cone_v, cone_n),
        ],
        [cap_f, cone_f],
    )
    return v, n, f


def create_arrow(
    radius: float = 0.1, height: float = 0.8, tip_radius: float = 0.2, tip_height: float = 0.2, sectors: int = 8
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    bot_cap_v, cap_n, cap_f = create_disk(radius, sectors)
    top_cap_v, _, _ = create_disk(tip_radius, sectors)

    cap_n = -cap_n
    cap_f = cap_f.copy().reshape((-1, 3))
    cap_f[:, (2, 1)] = cap_f[:, (1, 2)]
    cap_f = cap_f.flatten()
    top_cap_v = top_cap_v + np.array([0.0, 0.0, height], np.float32)

    cylinder_v, cylinder_n, cylinder_f = create_cylinder(radius, height, sectors)
    tip_v, tip_n, tip_f = create_cone(tip_radius, tip_height, sectors)
    tip_v[:, 2] += height
    (v, n), f = concatenate_meshes(
        [
            (bot_cap_v, cap_n),
            (top_cap_v, cap_n),
            (cylinder_v, cylinder_n),
            (tip_v, tip_n),
        ],
        [cap_f, cap_f, cylinder_f, tip_f],
    )
    return v, n, f


def create_plane(
    position: Tuple[float, float, float] = (0, 0, 0), extents: Tuple[float, float] = (1, 1)
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    v = np.array(
        [
            [-0.5, -0.5, 0.0],
            [0.5, -0.5, 0.0],
            [0.5, 0.5, 0.0],
            [-0.5, 0.5, 0.0],
        ],
        np.float32,
    ) * np.array([extents[0], extents[1], 0.0], np.float32) + np.array(position, np.float32)

    n = np.array(
        [
            [0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0],
        ],
        np.float32,
    )

    f = np.array([0, 1, 2, 0, 2, 3], np.uint32)

    return v, n, f


_cube_positions = np.array(
    [
        [-0.5, -0.5, -0.5],
        [0.5, -0.5, -0.5],
        [0.5, 0.5, -0.5],
        [-0.5, 0.5, -0.5],
        [-0.5, -0.5, 0.5],
        [0.5, -0.5, 0.5],
        [0.5, 0.5, 0.5],
        [-0.5, 0.5, 0.5],
        [-0.5, 0.5, -0.5],
        [-0.5, -0.5, -0.5],
        [-0.5, -0.5, 0.5],
        [-0.5, 0.5, 0.5],
        [0.5, -0.5, -0.5],
        [0.5, 0.5, -0.5],
        [0.5, 0.5, 0.5],
        [0.5, -0.5, 0.5],
        [-0.5, -0.5, -0.5],
        [0.5, -0.5, -0.5],
        [0.5, -0.5, 0.5],
        [-0.5, -0.5, 0.5],
        [0.5, 0.5, -0.5],
        [-0.5, 0.5, -0.5],
        [-0.5, 0.5, 0.5],
        [0.5, 0.5, 0.5],
    ],
    np.float32,
)

_cube_normals = np.array(
    [
        [0.0, 0.0, -1.0],
        [0.0, 0.0, -1.0],
        [0.0, 0.0, -1.0],
        [0.0, 0.0, -1.0],
        [0.0, 0.0, 1.0],
        [0.0, 0.0, 1.0],
        [0.0, 0.0, 1.0],
        [0.0, 0.0, 1.0],
        [-1.0, 0.0, 0.0],
        [-1.0, 0.0, 0.0],
        [-1.0, 0.0, 0.0],
        [-1.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, -1.0, 0.0],
        [0.0, -1.0, 0.0],
        [0.0, -1.0, 0.0],
        [0.0, -1.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 1.0, 0.0],
    ],
    np.float32,
)

_cube_indices = np.array(
    [
        #  front and back
        0,
        3,
        2,
        2,
        1,
        0,
        4,
        5,
        6,
        6,
        7,
        4,
        #  left and right
        11,
        8,
        9,
        9,
        10,
        11,
        12,
        13,
        14,
        14,
        15,
        12,
        #  bottom and top
        16,
        17,
        18,
        18,
        19,
        16,
        20,
        21,
        22,
        22,
        23,
        20,
    ],
    np.uint32,
)


def create_cube(
    position: Tuple[float, float, float] = (0, 0, 0), extents: Tuple[float, float, float] = (1, 1, 1)
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.uint32]]:
    return (
        _cube_positions * np.asarray(extents, np.float32) + np.asarray(position, np.float32),
        _cube_normals.copy(),
        _cube_indices.copy(),
    )


def create_cube_edges(
    min_p: Tuple[float, float, float] = (-0.5, -0.5, -0.5),
    max_p: Tuple[float, float, float] = (0.5, 0.5, 0.5),
    color: int = 0xFFFF0000,
) -> Tuple[NDArray[np.float32], NDArray[np.uint32]]:
    vertices = np.zeros((8, 3), dtype=np.float32)
    vertices[0:4] = min_p
    vertices[1, 0] = max_p[0]
    vertices[2, 0:2] = max_p[0:2]
    vertices[3, 1] = max_p[1]
    vertices[4:] = max_p
    vertices[4, 0:2] = min_p[0:2]
    vertices[7, 0] = min_p[0]
    vertices[5, 1] = min_p[1]

    lines = np.zeros((24, 3), dtype=np.float32)
    lines[0:2] = vertices[0:2]
    lines[2:4] = vertices[1:3]
    lines[4:6] = vertices[2:4]
    lines[6:8] = vertices[[3, 0]]
    lines[8:10] = vertices[4:6]
    lines[10:12] = vertices[5:7]
    lines[12:14] = vertices[6:8]
    lines[14:16] = vertices[[7, 4]]
    lines[16:18] = vertices[[0, 4]]
    lines[18:20] = vertices[[1, 5]]
    lines[20:22] = vertices[[2, 6]]
    lines[22:24] = vertices[[3, 7]]

    return lines, np.full(lines.shape[0], color, np.uint32)


def create_normal_lines(
    positions: NDArray[np.float32],
    normals: NDArray[np.float32],
    length: float = 0.01,
    color: int = 0xFFFF0000,
) -> Tuple[NDArray[np.float32], NDArray[np.uint32]]:
    n = positions.shape[0]
    lines = np.zeros((n * 2, 3), np.float32)
    lines[::2] = positions
    lines[1::2] = positions + normals * length
    return lines, np.full(lines.shape[0], color, np.uint32)


def create_axis3d_lines_and_colors(
    length: float = 1.0,
    x_color: int = 0xFF0000FF,
    y_color: int = 0xFF00FF00,
    z_color: int = 0xFFFF0000,
) -> Tuple[NDArray[np.float32], NDArray[np.uint32]]:
    lines = (
        np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0],
            ],
            np.float32,
        )
        * length
    )

    colors = np.array(
        [
            x_color,
            x_color,
            y_color,
            y_color,
            z_color,
            z_color,
        ],
        np.uint32,
    )

    return lines, colors


def create_axis2d_lines_and_colors(
    length: float = 1.0,
    x_color: int = 0xFF0000FF,
    y_color: int = 0xFF00FF00,
) -> Tuple[NDArray[np.float32], NDArray[np.uint32]]:
    lines = (
        np.array(
            [
                [0.0, 0.0],
                [1.0, 0.0],
                [0.0, 0.0],
                [0.0, 1.0],
            ],
            np.float32,
        )
        * length
    )

    colors = np.array(
        [
            x_color,
            x_color,
            y_color,
            y_color,
        ],
        np.uint32,
    )

    return lines, colors


def concatenate_meshes(
    attributes: Sequence[Sequence[NDArray[Any]]], indices: Sequence[NDArray[np.uint32]]
) -> Tuple[List[NDArray[np.float32]], NDArray[np.uint32]]:
    total_offset = 0
    offsets = []
    for m in attributes:
        a_len = m[0].shape[0]
        for a in m[1:]:
            if a_len != a.shape[0]:
                raise ValueError("all attributes must have the same number of vertices")
        offsets.append(total_offset)
        total_offset += a_len

    cc_attributes = [
        np.concatenate([attributes[m][i] for m in range(len(attributes))]) for i in range(len(attributes[0]))
    ]
    cc_indices = np.concatenate([idx + o for idx, o in zip(indices, offsets)])
    return cc_attributes, cc_indices


def transform_positions(transform: mat4, vertices: NDArray[np.float32]) -> NDArray[np.float32]:
    t = np.array(transform)
    v_h = np.hstack((vertices, np.ones((vertices.shape[0], 1), np.float32)))
    return (v_h @ t.T)[:, :3]  # type: ignore


def transform_directions(transform: mat3, directions: NDArray[np.float32]) -> NDArray[np.float32]:
    t = np.array(transform)
    return directions @ t.T  # type: ignore


def transform_mesh(
    transform: mat4, vertices: NDArray[np.float32], normals: NDArray[np.float32]
) -> Tuple[NDArray[np.float32], NDArray[np.float32]]:
    return transform_positions(transform, vertices), transform_directions(mat3(transform), normals)



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

import hashlib
import logging
import pickle
import shutil
from pathlib import Path
from typing import List, Optional, Tuple, Union

from platformdirs import user_cache_path
from pyxpg import slang

logger = logging.getLogger(__name__)

CACHE_DIR = user_cache_path("ambra")

CACHE_VERSION_MAJOR = 0
CACHE_VERSION_MINOR = 3
CACHE_VERSION_PATCH = 0
CACHE_VERSION = f"{CACHE_VERSION_MAJOR}.{CACHE_VERSION_MINOR}.{CACHE_VERSION_PATCH}"


def clear_cache() -> None:
    shutil.rmtree(CACHE_DIR, ignore_errors=True)


def vulkan_version_to_minimum_supported_spirv_version(
    version: Tuple[int, int],
) -> str:
    if version[0] == 0:
        raise ValueError("Unsupported Vulkan version < 1")
    if version[0] == 1:
        if version[1] == 0:
            return "spirv_1_0"
        elif version[1] == 1:
            return "spirv_1_3"
        elif version[1] == 2:
            return "spirv_1_5"
        else:
            return "spirv_1_6"
    else:
        return "spirv_1_6"


def compile(
    file: Path,
    entry: str = "main",
    target: str = "spirv_1_3",
    defines: Optional[List[Tuple[str, str]]] = None,
    include_paths: Optional[List[Union[Path, str]]] = None,
) -> slang.Shader:
    defines_list = defines or []
    include_paths_list = [str(p) for p in include_paths] if include_paths is not None else []

    defines_bytes = b"".join([f"{k}\0{v}\0".encode() for k, v in sorted(defines_list)])
    include_paths_bytes = b"".join([f"{p}\0".encode() for p in sorted(include_paths_list)])

    hexdigest = hashlib.sha256(b"\0".join([defines_bytes, include_paths_bytes, file.read_bytes()])).hexdigest()
    name = f"{hexdigest}_{entry}_{target}_{CACHE_VERSION}.shdr"

    # Check cache
    path = Path(CACHE_DIR, name)
    if path.exists():
        try:
            with path.open("rb") as f:
                pkl = pickle.load(f)
            prog: slang.Shader = pkl[0]
            old_hashes: List[str] = pkl[1]

            # Check if any of the dependent files changed from when the shader
            # was serialized.
            assert len(prog.dependencies) == len(old_hashes)
            new_hashes = [hashlib.sha256(Path(d).read_bytes()).hexdigest() for d in sorted(prog.dependencies)]
            if new_hashes == old_hashes:
                logger.info("Shader cache hit: %s -> %s ", file, path)
                return prog
        except Exception as e:
            logger.warning("Shader cache hit invalid: %s (%s)", file, e)

    # Create prog
    logger.info("Shader cache miss: %s", file)
    prog = slang.compile(str(file), entry, target, defines_list, include_paths_list)
    hashes = [hashlib.sha256(Path(d).read_bytes()).hexdigest() for d in sorted(prog.dependencies)]

    # Populate cache
    CACHE_DIR.mkdir(parents=True, exist_ok=True)
    with path.open("wb") as f:
        pickle.dump((prog, hashes), f)
    return prog



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from collections import OrderedDict
from dataclasses import dataclass
from typing import Callable, Dict, Generic, List, Optional, Sequence, Tuple, TypeVar


class _RefCount:
    def __init__(self, initial_count: int = 0):
        self.count = initial_count

    def inc(self) -> None:
        self.count += 1

    def dec(self) -> bool:
        self.count -= 1
        return self.count == 0

    def __repr__(self) -> str:
        return f"RC({self.count})"


K = TypeVar("K")
O = TypeVar("O")


@dataclass
class _Entry(Generic[O]):
    obj: O
    refcount: _RefCount
    prefetching: bool


class LRUPool(Generic[K, O]):
    def __init__(
        self,
        objs: List[O],
        num_frames: int,
        max_prefetch: int = 0,
    ):
        # (Key, Generation) -> Entry: holds all entries in use
        self.lookup: OrderedDict[Tuple[K, int], _Entry[O]] = OrderedDict()

        # Object -> (Key, Generation): holds all free objects. If the object
        # is not yet associated with any key (if it was never used before)
        # the value for that item will be None.
        self.lru: OrderedDict[O, Optional[Tuple[K, int]]] = OrderedDict()

        # ((Key, Generation), Entry): list of entries in flight with their key and generation
        self.in_flight: List[Optional[Tuple[Tuple[K, int], _Entry[O]]]] = [None] * num_frames

        # Key -> Generation: current generation of in use keys. If a key is not
        # in use, it will not be in this map and its generation will be implicitly 0.
        #
        # The generation index is increased when requested by the user.
        # It will also be reset when there is no entry associated with that
        # key in the lookup anymore (this implies no reference also in the lru,
        # in_flight list, and prefetch_store).
        self.current_generation: Dict[K, int] = {}

        # Maximum number of prefetched items at a time
        self.max_prefetch: int = max_prefetch

        # Object -> (Key, Generation): stores state for each prefetching item
        self.prefetch_store: Dict[O, Tuple[K, int]] = {}

        # Initialize LRU to empty objects
        for o in objs:
            self.lru[o] = None

    def get(
        self,
        k: K,
        load: Callable[[K, O], None],
        ensure_fetched: Optional[Callable[[K, O], None]] = None,
    ) -> O:
        key = (k, self.current_generation.get(k, 0))
        cached = self.lookup.get(key)

        # Check if already loaded
        if cached is None:
            # Grab a free object from the lru
            obj = self._lru_pop_free()

            # Load
            load(key[0], obj)

            # Register buffer as loaded for future use
            self.lookup[key] = _Entry(obj, _RefCount(), prefetching=False)
        else:
            # If this was a prefetched buffer wait for it to be loaded
            obj = cached.obj
            if cached.prefetching:
                # Realize prefetch
                if ensure_fetched:
                    ensure_fetched(key[0], obj)

                # Item is still in the prefetching list here.
                # It will be removed by the next prefetch cleanup
                cached.prefetching = False

            # If the buffer is in the LRU remove it to mark it as in use
            try:
                self.lru.pop(obj)
            except KeyError:
                pass
        return obj

    def is_available(self, k: K) -> bool:
        cached = self.lookup.get((k, self.current_generation.get(k, 0)))
        return cached is not None and not cached.prefetching

    def is_available_or_prefetching(self, k: K) -> bool:
        cached = self.lookup.get((k, self.current_generation.get(k, 0)))
        return cached is not None

    def increment_generation(self, k: K) -> None:
        gen = self.current_generation.get(k, 0)
        if (o := self.lookup.get((k, gen))) is not None:
            self.current_generation[k] = gen + 1
            try:
                self.lru.move_to_end(o.obj, last=False)
            except KeyError:
                pass

    def increment_all_generations(self) -> None:
        for k, o in self.lookup.items():
            current_gen = self.current_generation.get(k[0], 0)
            if current_gen == k[1]:
                self.current_generation[k[0]] = k[1] + 1
            try:
                self.lru.move_to_end(o.obj, last=False)
            except KeyError:
                pass

    def use_frame(self, frame_index: int, k: K) -> None:
        key = (k, self.current_generation.get(k, 0))

        entry = self.lookup[key]
        entry.refcount.inc()
        self.in_flight[frame_index] = (key, entry)

    def use_manual(self, k: K) -> None:
        key = (k, self.current_generation.get(k, 0))

        entry = self.lookup[key]
        entry.refcount.inc()

    def release_manual(self, k: K) -> None:
        key = (k, self.current_generation.get(k, 0))

        entry = self.lookup[key]

        # Decrement refcount
        if entry.refcount.dec():
            # If refcount is 0 add buffer back to LRU
            self.lru[entry.obj] = key

            # If this is an old generation of this buffer move it
            # to the end of the LRU to make sure it will be evicted soon.
            if key[1] < self.current_generation.get(key[0], 0):
                self.lru.move_to_end(entry.obj, last=False)

    def release_frame(self, frame_index: int) -> None:
        if old := self.in_flight[frame_index]:
            key, entry = old

            # Decrement refcount
            if entry.refcount.dec():
                # If refcount is 0 add buffer back to LRU
                self.lru[entry.obj] = key

                # If this is an old generation of this buffer move it
                # to the end of the LRU to make sure it will be evicted soon.
                if key[1] < self.current_generation.get(key[0], 0):
                    self.lru.move_to_end(entry.obj, last=False)

            # Mark nothing in flight yet for this frame.
            self.in_flight[frame_index] = None

    def give_back(self, k: K, obj: O) -> None:
        self.lru[obj] = (k, self.current_generation.get(k, 0))

    def prefetch(
        self,
        useful_range: Sequence[K],
        cleanup: Callable[[K, O], bool],
        submit_load: Callable[[K, O], None],
    ) -> None:
        if self.max_prefetch <= 0:
            return

        for obj, key in list(self.prefetch_store.items()):
            if cleanup(key[0], obj):
                # Insert back in the LRU if not yet claimed
                if self.lookup[key].prefetching:
                    self.lru[obj] = key

                    # Check if the buffer is old or it's not in the window
                    if key[1] < self.current_generation.get(key[0], 0) or key[0] not in useful_range:
                        # Move to end of LRU to ensure this buffer is reused soon
                        self.lru.move_to_end(obj, last=False)

                    # Mark as ready
                    self.lookup[key].prefetching = False

                # Remove from prefetching dict
                del self.prefetch_store[obj]

            # TODO: we could also cancel prefetch work here, if possible

        bump = []
        prefetch_count = self.max_prefetch - len(self.prefetch_store)

        # Submit prefetch requests and allocate buffers in ascending distance.
        for k in useful_range[:prefetch_count]:
            key = (k, self.current_generation.get(k, 0))
            cached = self.lookup.get(key)
            if cached is None:
                # Grab a free object from the lru
                obj = self._lru_pop_free()

                # Submit for load
                submit_load(key[0], obj)

                # Register buffer as loading for future use
                self.lookup[key] = _Entry(obj, _RefCount(), prefetching=True)
                self.prefetch_store[obj] = key
            else:
                # If already loaded just bump in front of LRU
                bump.append(cached.obj)

        # Bump to front in reverse order to ensure the closest element
        # will be evicted as late as possible.
        for o in reversed(bump):
            try:
                # Refresh entry in LRU cache
                self.lru.move_to_end(o)
            except KeyError:  # noqa: PERF203
                pass

    def clear(self) -> None:
        self.lru.clear()
        self.lookup.clear()
        self.in_flight.clear()
        self.prefetch_store.clear()
        self.current_generation.clear()

    def _lru_pop_free(self) -> O:
        obj, old_key = self.lru.popitem(last=False)
        if old_key is not None:
            # Remove the bufer from the lookup
            self.lookup.pop(old_key)

            # If we are removing the buffer with the current generation,
            # reset its generation number to 0 by removing it.
            if self.current_generation.get(old_key[0]) == old_key[1]:
                del self.current_generation[old_key[0]]
        return obj



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import Enum, Flag, auto
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, Set, Tuple, TypeVar, Union

from numpy.typing import NDArray
from pyxpg import (
    AccessFlags,
    AllocType,
    Buffer,
    BufferBarrier,
    BufferUsageFlags,
    CommandBuffer,
    DescriptorSet,
    DescriptorType,
    Device,
    Format,
    Image,
    ImageBarrier,
    ImageCreateFlags,
    ImageLayout,
    ImageUsageFlags,
    ImageView,
    ImageViewType,
    MemoryUsage,
    PipelineStageFlags,
    TimelineSemaphore,
)

from .config import UploadMethod
from .ffx import MipGenerationRequest, SPDPipelineInstance
from .renderer_frame import RendererFrame, SemaphoreInfo
from .utils.gpu import (
    BufferUploadInfo,
    ImageUploadInfo,
    MipGenerationFilter,
    get_image_pitch_rows_and_texel_size,
    to_srgb_format,
    view_bytes,
)
from .utils.lru_pool import LRUPool
from .utils.ring_buffer import RingBuffer
from .utils.threadpool import Promise

if TYPE_CHECKING:
    from .property import BufferProperty, ImageProperty, Property, PropertyItem
    from .renderer import Renderer


class CpuBuffer:
    def __init__(self, buffer: Buffer):
        self.buf = buffer
        self.used_size = 0
        self.promise = Promise[None]()

    def __repr__(self) -> str:
        return self.buf.__repr__()

    def destroy(self) -> None:
        self.buf.destroy()


class GpuResourceState(Enum):
    EMPTY = auto()
    LOAD = auto()
    PREFETCH = auto()
    RENDER = auto()


@dataclass
class GpuBufferView:
    buffer: Buffer
    offset: int
    size: int

    def destroy(self) -> None:
        self.buffer.destroy()

    def buffer_and_offset(self) -> Tuple[Buffer, int]:
        return (self.buffer, self.offset)

    def write_descriptor(
        self, descriptor_set: DescriptorSet, type: DescriptorType, binding: int, element: int = 0
    ) -> None:
        descriptor_set.write_buffer(self.buffer, type, binding, element, self.offset, self.size)

    def append_view_to_destroy_list(self, destroy_list: List[Union[Buffer, Image, ImageView]]) -> None:
        pass

    def resource(self) -> Buffer:
        return self.buffer


@dataclass
class GpuImageView:
    image: Image
    srgb_resource_view: Optional[ImageView]
    mip_level_0_view: Optional[ImageView]
    mip_views: List[ImageView]

    def destroy(self) -> None:
        self.image.destroy()

    def view(self) -> Union[Image, ImageView]:
        if self.srgb_resource_view is not None:
            return self.srgb_resource_view
        else:
            return self.image

    def append_view_to_destroy_list(self, destroy_list: List[Union[Buffer, Image, ImageView]]) -> None:
        if self.srgb_resource_view is not None:
            destroy_list.append(self.srgb_resource_view)
        if self.mip_level_0_view is not None:
            destroy_list.append(self.mip_level_0_view)
        destroy_list.extend(self.mip_views)

    def resource(self) -> Image:
        return self.image


def _create_image_view(
    device: Device,
    width: int,
    height: int,
    depth: int,
    format: Format,
    usage_flags: ImageUsageFlags,
    srgb: bool,
    mips: bool,
    name: str,
) -> GpuImageView:
    image_create_flags = ImageCreateFlags.NONE
    if srgb:
        image_create_flags |= ImageCreateFlags.MUTABLE_FORMAT

    mip_levels = 1
    if mips:
        assert depth == 1
        mip_levels = max(width.bit_length(), height.bit_length())
        usage_flags |= ImageUsageFlags.STORAGE

    img = Image(
        device,
        width,
        height,
        format,
        usage_flags,
        AllocType.DEVICE,
        depth=depth,
        mip_levels=mip_levels,
        create_flags=image_create_flags,
        name=name,
    )
    srgb_view = None
    if srgb:
        srgb_view = ImageView(
            device,
            img,
            ImageViewType.TYPE_2D if depth == 1 else ImageViewType.TYPE_3D,
            to_srgb_format(format),
            usage_flags=ImageUsageFlags.SAMPLED,
            name=f"{name}-srgb",
        )

    mip_level_0_view = None
    mip_views = []
    if mips:
        mip_level_0_view = ImageView(
            device,
            img,
            ImageViewType.TYPE_2D_ARRAY,
            to_srgb_format(format) if srgb else format,
            usage_flags=ImageUsageFlags.SAMPLED,
            name=f"{name}-mip-0",
        )
        mip_views = [
            ImageView(
                device,
                img,
                ImageViewType.TYPE_2D_ARRAY,
                base_mip_level=m if m < img.mip_levels else 0,
                mip_level_count=1,
                name=f"{name}-mip-{m + 1}",
            )
            for m in range(mip_levels)
        ]

    return GpuImageView(img, srgb_view, mip_level_0_view, mip_views)


R = TypeVar("R", bound=Union[Buffer, Image])
V = TypeVar("V", bound=Union[GpuBufferView, GpuImageView])


class GpuResource(Generic[V]):
    def __init__(self, resource: V, semaphore: Optional[TimelineSemaphore]):
        self.resource = resource
        self.state = GpuResourceState.EMPTY
        self.semaphore = semaphore
        self.semaphore_value = 0

    def use(self, stage: PipelineStageFlags) -> SemaphoreInfo:
        assert self.semaphore is not None
        info = SemaphoreInfo(
            self.semaphore,
            stage,
            self.semaphore_value,
            stage,
            self.semaphore_value + 1,
        )
        self.semaphore_value += 1
        return info

    def __repr__(self) -> str:
        return f"(resource={self.resource.__repr__()}, state={self.state}, semaphore={self.semaphore_value})"

    def destroy(self) -> None:
        if self.semaphore is not None:
            self.semaphore.destroy()
        self.resource.destroy()


@dataclass
class PrefetchState:
    commands: CommandBuffer
    prefetch_done_value: int


class GpuPropertySupportedOperations(Flag):
    UPDATE = auto()
    APPEND = auto()
    POP = auto()
    INSERT = auto()
    REMOVE = auto()


class GpuProperty(Generic[V]):
    supported_operations: GpuPropertySupportedOperations
    name: str

    # Public API
    def get_current(self) -> V:
        raise NotImplementedError

    def enqueue_for_destruction(self) -> None:
        pass

    # Renderer API
    def load(self, frame: RendererFrame) -> None:
        pass

    def upload(self, frame: RendererFrame) -> None:
        pass

    def prefetch(self) -> None:
        pass

    # Edit API
    def update_frame(self, index: int) -> None:
        raise NotImplementedError

    def update_frames(self, indices: List[int]) -> None:
        raise NotImplementedError

    def update_frame_range(self, start: int, stop: int) -> None:
        raise NotImplementedError

    def append_frame(self) -> None:
        raise NotImplementedError

    def append_frames(self, count: int) -> None:
        raise NotImplementedError

    def pop_frame(self) -> None:
        raise NotImplementedError

    def pop_frames(self, count: int) -> None:
        raise NotImplementedError

    def insert_frame(self, index: int) -> None:
        raise NotImplementedError

    def insert_frames(self, index: int, count: int) -> None:
        raise NotImplementedError

    def remove_frame(self, index: int) -> None:
        raise NotImplementedError

    def remove_frames(self, indices: List[int]) -> None:
        raise NotImplementedError

    def remove_frame_range(self, start: int, stop: int) -> None:
        raise NotImplementedError


class GpuPreuploadedArrayProperty(GpuProperty[V], Generic[R, V]):
    def __init__(
        self,
        data: NDArray[Any],
        frame_size: int,
        renderer: "Renderer",
        upload_method: UploadMethod,
        property: "Property",
        name: str,
    ):
        self.frame_size = frame_size
        self.renderer = renderer
        self.upload_method = upload_method
        self.property = property
        self.supported_operations = GpuPropertySupportedOperations.UPDATE
        self.name = name

        self.resource, self.resource_views = self._create_preuploaded_array_resource_and_views(
            data, renderer.bulk_upload_list
        )
        self.staging_buffers: Optional[RingBuffer[Buffer]] = None
        self.current_staging_buf: Optional[Buffer] = None
        self.invalid_frames: Set[int] = set()
        self.current: Optional[V] = None

    # Public API
    def get_current(self) -> V:
        assert self.current is not None
        return self.current

    def enqueue_for_destruction(self) -> None:
        resources: List[Union[Buffer, Image, ImageView]] = [self.resource]
        if self.staging_buffers is not None:
            resources.extend(self.staging_buffers.items)
            self.staging_buffers = None

    def load(self, frame: RendererFrame) -> None:
        index = self.property.current_frame_index
        if index in self.invalid_frames:
            self.invalid_frames.remove(index)

            assert self.staging_buffers is not None

            buf = self.staging_buffers.get_current_and_advance()
            self.property.get_frame_by_index_into(index, buf.data)
            self.current_staging_buf = buf

            self._append_barriers_and_mip_requests_for_upload(frame, self.resource_views[index])

    # Renderer API
    def upload(self, frame: RendererFrame) -> None:
        view = self.resource_views[self.property.current_frame_index]
        if self.current_staging_buf is not None:
            self._cmd_upload(frame.cmd, self.current_staging_buf, view)
            self.current_staging_buf = None
        self.current = view

    # Edit API
    def update_frame(self, index: int) -> None:
        self.invalid_frames.add(index)
        self._promote_to_dynamic()

    def update_frames(self, indices: List[int]) -> None:
        for i in indices:
            self.invalid_frames.add(i)
        self._promote_to_dynamic()

    def update_frame_range(self, start: int, stop: int) -> None:
        for i in range(start, stop):
            self.invalid_frames.add(i)
        self._promote_to_dynamic()

    # Private API
    def _promote_to_dynamic(self) -> None:
        if self.staging_buffers is None:
            self.staging_buffers = RingBuffer(
                [
                    Buffer(
                        self.renderer.device,
                        self.frame_size,
                        BufferUsageFlags.TRANSFER_SRC,
                        AllocType.HOST,
                        f"{self.name}-staging-{i}",
                    )
                    for i in range(self.renderer.num_frames_in_flight)
                ]
            )

    def _create_preuploaded_array_resource_and_views(
        self, data: NDArray[Any], out_upload_list: List[Union[BufferUploadInfo, ImageUploadInfo]]
    ) -> Tuple[R, List[V]]:
        raise NotImplementedError

    def _cmd_upload(self, cmd: CommandBuffer, staging_buffer: Buffer, view: V) -> None:
        raise NotImplementedError

    def _append_barriers_and_mip_requests_for_upload(self, frame: RendererFrame, resource: V) -> None:
        raise NotImplementedError


class GpuBufferPreuploadedArrayProperty(GpuPreuploadedArrayProperty[Buffer, GpuBufferView]):
    def __init__(
        self,
        data: NDArray[Any],
        renderer: "Renderer",
        property: "BufferProperty",
        name: str,
    ):
        self.usage_flags = property.gpu_usage | BufferUsageFlags.TRANSFER_DST
        self.pipeline_stage_flags = property.gpu_stage

        super().__init__(
            data,
            property.max_size,
            renderer,
            renderer.buffer_upload_method,
            property,
            name,
        )

    # Private API
    def _create_preuploaded_array_resource_and_views(
        self, data: NDArray[Any], out_upload_list: List[Union[BufferUploadInfo, ImageUploadInfo]]
    ) -> Tuple[Buffer, List[GpuBufferView]]:
        if self.upload_method == UploadMethod.MAPPED_PREFER_HOST:
            alloc_type = AllocType.HOST
        elif self.upload_method == UploadMethod.MAPPED_PREFER_DEVICE:
            alloc_type = AllocType.DEVICE_MAPPED
        else:
            alloc_type = AllocType.DEVICE

        # TODO: handle alignment here (need to pad the data)
        buffer = Buffer(self.renderer.device, data.nbytes, self.usage_flags, alloc_type, self.name)

        if buffer.is_mapped:
            buffer.data[:] = view_bytes(data)
        else:
            out_upload_list.append(BufferUploadInfo(view_bytes(data), buffer, 0))

        views = [GpuBufferView(buffer, self.frame_size * i, self.frame_size) for i in range(self.property.num_frames)]

        return buffer, views

    def _cmd_upload(self, cmd: CommandBuffer, staging_buffer: Buffer, view: GpuBufferView) -> None:
        cmd.copy_buffer_range(staging_buffer, view.buffer, self.frame_size, 0, view.offset)

    def _append_barriers_and_mip_requests_for_upload(self, frame: RendererFrame, resource: V) -> None:
        frame.upload_property_pipeline_stages |= self.pipeline_stage_flags


class GpuPreuploadedProperty(GpuProperty[V], Generic[R, V]):
    def __init__(
        self,
        max_frame_size: int,
        batched: bool,
        renderer: "Renderer",
        upload_method: UploadMethod,
        property: "Property",
        name: str,
    ):
        self.max_frame_size = max_frame_size
        self.batched = batched
        self.renderer = renderer
        self.upload_method = upload_method
        self.property = property
        self.supported_operations = (
            GpuPropertySupportedOperations.UPDATE
            | GpuPropertySupportedOperations.APPEND
            | GpuPropertySupportedOperations.POP
            | GpuPropertySupportedOperations.INSERT
            | GpuPropertySupportedOperations.REMOVE
            if not self.batched
            else GpuPropertySupportedOperations.UPDATE
        )
        self.name = name

        self.async_load = self.property.upload.async_load
        self.staging_buffers: Optional[RingBuffer[CpuBuffer]] = None
        self.current_staging_buf: Optional[CpuBuffer] = None
        self.invalid_frames: Set[int] = set()
        self.current: Optional[V] = None

        # NOTE: In the async_load case we could also do mapped upload in
        # threads here, but we first need to know the size of the frame.
        # This makes things more complicated because we could run into
        # issues using the context from multiple threads. It's not
        # clear where is the best way to ensure this is thread safe so
        # for now we don't do it.
        if self.async_load:
            promises: List[Promise[PropertyItem]] = []
            for i in range(property.num_frames):
                promise: Promise[PropertyItem] = Promise()
                self.renderer.thread_pool.submit(promise, self._load_async, i)  # type: ignore
                promises.append(promise)

        # Collect frames
        frames = []
        for i in range(property.num_frames):
            if self.async_load:
                frame = promises[i].get()
            else:
                frame = property.get_frame_by_index(i)
            frames.append(frame)

        # Create resource and views
        self.resources, self.resource_views = self._create_preuploaded_resources_and_views(frames)

    # Public API
    def get_current(self) -> V:
        assert self.current is not None
        return self.current

    def enqueue_for_destruction(self) -> None:
        destroy_list: List[Union[Buffer, Image, ImageView]] = self.resources.copy()  # type: ignore
        for v in self.resource_views:
            v.append_view_to_destroy_list(destroy_list)
        self.renderer.enqueue_for_destruction(destroy_list)

    # Renderer API
    def load(self, frame: RendererFrame) -> None:
        index = self.property.current_frame_index
        if index in self.invalid_frames:
            self.invalid_frames.remove(index)

            assert self.staging_buffers is not None

            buf = self.staging_buffers.get_current_and_advance()
            if self.async_load:
                self.renderer.thread_pool.submit(buf.promise, self._load_async_into, index, buf)  # type: ignore
            else:
                buf.used_size = self.property.get_frame_by_index_into(index, buf.buf.data)
            self.current_staging_buf = buf

            self._append_barriers_and_mip_requests_for_upload(frame, self.resource_views[index])

    def upload(self, frame: RendererFrame) -> None:
        view = self.resource_views[self.property.current_frame_index]
        if self.current_staging_buf is not None:
            # Wait for buffer to be ready
            if self.async_load:
                self.current_staging_buf.promise.get()

            self._cmd_upload(frame.cmd, self.current_staging_buf.buf, view)
            self.current_staging_buf = None
        self.current = view

    # Edit API
    def update_frame(self, index: int) -> None:
        self.invalid_frames.add(index)
        self._promote_to_dynamic()

    def update_frames(self, indices: List[int]) -> None:
        for i in indices:
            self.invalid_frames.add(i)
        self._promote_to_dynamic()

    def update_frame_range(self, start: int, stop: int) -> None:
        for i in range(start, stop):
            self.invalid_frames.add(i)
        self._promote_to_dynamic()

    def append_frame(self) -> None:
        if self.batched:
            raise RuntimeError("Appending to a preuploaded and batched GPU property is not supported.")
        self.append_frames(1)

    def append_frames(self, count: int) -> None:
        if self.batched:
            raise RuntimeError("Appending to a preuploaded and batched GPU property is not supported.")

        current_count = len(self.resource_views)
        new_max_frame_size = 0
        for i in range(count):
            new_frame = self.property.get_frame_by_index(current_count + i)
            new_max_frame_size = max(new_max_frame_size, len(view_bytes(new_frame)))

            # Create and append a new resource and views
            resource, view = self._create_preuploaded_resource_and_view_for_frame(new_frame)
            self.resources.append(resource)
            self.resource_views.append(view)

        self._update_max_frame_size(new_max_frame_size)

    def pop_frame(self) -> None:
        if self.batched:
            raise RuntimeError("Popping from a preuploaded and batched GPU property is not supported.")

        res = self.resources.pop()
        view = self.resource_views.pop()

        destroy_list: List[Union[Buffer, Image, ImageView]] = [res]
        view.append_view_to_destroy_list(destroy_list)

        self.renderer.enqueue_for_destruction(destroy_list)

    def pop_frames(self, count: int) -> None:
        if self.batched:
            raise RuntimeError("Popping from a preuploaded and batched GPU property is not supported.")

        if count > 0:
            destroy_list: List[Union[Buffer, Image, ImageView]] = self.resources[-count:]  # type: ignore
            for v in self.resource_views[-count:]:
                v.append_view_to_destroy_list(destroy_list)
            self.renderer.enqueue_for_destruction(destroy_list)

            del self.resources[-count:]
            del self.resource_views[-count:]

    def insert_frame(self, index: int) -> None:
        if self.batched:
            raise RuntimeError("Inserting into a preuploaded and batched GPU property is not supported.")

        new_frame = self.property.get_frame_by_index(index)
        self._update_max_frame_size(len(view_bytes(new_frame)))

        # Create and append a new resource and views
        resource, view = self._create_preuploaded_resource_and_view_for_frame(new_frame)
        self.resources.insert(index, resource)
        self.resource_views.insert(index, view)

    def insert_frames(self, index: int, count: int) -> None:
        if self.batched:
            raise RuntimeError("Inserting in preuploaded and batched GPU property is not supported.")

        current_count = len(self.resource_views)

        new_resources = []
        new_resource_views = []
        new_max_frame_size = 0
        for i in range(count):
            new_frame = self.property.get_frame_by_index(current_count + i)
            new_max_frame_size = max(new_max_frame_size, len(view_bytes(new_frame)))

            # Create and append a new resource and views
            resource, view = self._create_preuploaded_resource_and_view_for_frame(new_frame)
            new_resources.append(resource)
            new_resource_views.append(view)

        self.resources = self.resources[:index] + new_resources + self.resources[index:]
        self.resource_views = self.resource_views[:index] + new_resource_views + self.resource_views[index:]

        self._update_max_frame_size(new_max_frame_size)

    def remove_frame(self, index: int) -> None:
        if self.batched:
            raise RuntimeError("Removing from a preuploaded and batched GPU property is not supported.")

        res = self.resources.pop(index)
        view = self.resource_views.pop(index)

        destroy_list: List[Union[Buffer, Image, ImageView]] = [res]
        view.append_view_to_destroy_list(destroy_list)
        self.renderer.enqueue_for_destruction([res])

    def remove_frames(self, indices: List[int]) -> None:
        if self.batched:
            raise RuntimeError("Removing from a preuploaded and batched GPU property is not supported.")

        indices_set = set(indices)

        keep_resources: List[R] = []
        keep_resource_views: List[V] = []
        destroy_list: List[Union[Buffer, Image, ImageView]] = []
        for i, (r, v) in enumerate(zip(self.resources, self.resource_views)):
            if i in indices_set:
                destroy_list.append(r)
                v.append_view_to_destroy_list(destroy_list)
            else:
                keep_resources.append(r)
                keep_resource_views.append(v)

        self.resources = keep_resources
        self.resource_views = keep_resource_views
        self.renderer.enqueue_for_destruction(destroy_list)

    def remove_frame_range(self, start: int, stop: int) -> None:
        if self.batched:
            raise RuntimeError("Removing from a preuploaded and batched GPU property is not supported.")

        # Collect resources to destroy and enqueue for destruction
        destroy_list: List[Union[Buffer, Image, ImageView]] = self.resources[start:stop]  # type: ignore
        for v in self.resource_views[start:stop]:
            v.append_view_to_destroy_list(destroy_list)
        self.renderer.enqueue_for_destruction(destroy_list)

        # Remove resources from lists
        self.resources = self.resources[:start] + self.resources[stop:]
        self.resource_views = self.resource_views[:start] + self.resource_views[stop:]

    # Private API
    def _promote_to_dynamic(self) -> None:
        if self.staging_buffers is None:
            self.staging_buffers = RingBuffer(
                [
                    CpuBuffer(
                        Buffer(
                            self.renderer.device,
                            self.max_frame_size,
                            BufferUsageFlags.TRANSFER_SRC,
                            AllocType.HOST,
                            f"{self.name}-staging-{i}",
                        )
                    )
                    for i in range(self.renderer.num_frames_in_flight)
                ]
            )

    def _update_max_frame_size(self, new_max_frame_size: int) -> None:
        # Update maximum frame size and clear current staging buffers if they can't fit the new frame.
        if new_max_frame_size > self.max_frame_size:
            self.max_frame_size = new_max_frame_size
            if self.staging_buffers is not None:
                self.renderer.enqueue_for_destruction([b.buf for b in self.staging_buffers])
            self.staging_buffers = None

    def _load_async(self, i: int, thread_index: int) -> "PropertyItem":
        return self.property.get_frame_by_index(i, thread_index)

    def _load_async_into(self, i: int, buf: CpuBuffer, thread_index: int) -> None:
        buf.used_size = self.property.get_frame_by_index_into(i, buf.buf.data, thread_index)

    def _create_preuploaded_resources_and_views(self, frames: List[NDArray[Any]]) -> Tuple[List[R], List[V]]:
        raise NotImplementedError

    def _create_preuploaded_resource_and_view_for_frame(self, frame: NDArray[Any]) -> Tuple[R, V]:
        raise NotImplementedError

    def _cmd_upload(self, cmd: CommandBuffer, staging_buffer: Buffer, view: V) -> None:
        raise NotImplementedError

    def _append_barriers_and_mip_requests_for_upload(self, frame: RendererFrame, resource: V) -> None:
        raise NotImplementedError


class GpuBufferPreuploadedProperty(GpuPreuploadedProperty[Buffer, GpuBufferView]):
    def __init__(
        self,
        renderer: "Renderer",
        property: "BufferProperty",
        name: str,
    ):
        self.usage_flags = property.gpu_usage | BufferUsageFlags.TRANSFER_DST
        self.pipeline_stage_flags = property.gpu_stage

        upload_method = renderer.buffer_upload_method
        if upload_method == UploadMethod.MAPPED_PREFER_HOST:
            self.alloc_type = AllocType.HOST
        elif upload_method == UploadMethod.MAPPED_PREFER_DEVICE:
            self.alloc_type = AllocType.DEVICE_MAPPED
        else:
            self.alloc_type = AllocType.DEVICE

        super().__init__(
            property.max_size,
            property.upload.batched,
            renderer,
            upload_method,
            property,
            name,
        )

    # Private API
    def _create_preuploaded_resources_and_views(
        self, frames: List[NDArray[Any]]
    ) -> Tuple[List[Buffer], List[GpuBufferView]]:
        resources: List[Buffer] = []
        views: List[GpuBufferView] = []
        if self.batched:
            # If batched allocate a single buffer for all the frames
            total_size_in_bytes = 0
            for frame in frames:
                total_size_in_bytes += len(view_bytes(frame))

            buffer = Buffer(self.renderer.device, total_size_in_bytes, self.usage_flags, self.alloc_type, self.name)
            resources.append(buffer)

            # Create views and upload
            offset = 0
            for frame in frames:
                frame_bytes = view_bytes(frame)
                size = len(frame_bytes)

                if buffer.is_mapped:
                    buffer.data[offset : offset + size] = frame_bytes
                else:
                    self.renderer.bulk_upload_list.append(BufferUploadInfo(frame_bytes, buffer, offset))
                views.append(GpuBufferView(buffer, offset, size))

                # TODO: handle alignment here depending on usage flags (vbo, ubo, ssbo alignement constraints)
                offset += size
        else:
            # If not batched allocate a separate buffer each frame
            for frame in frames:
                resource, view = self._create_preuploaded_resource_and_view_for_frame(frame)
                resources.append(resource)
                views.append(view)

        return resources, views

    def _create_preuploaded_resource_and_view_for_frame(self, frame: NDArray[Any]) -> Tuple[Buffer, GpuBufferView]:
        frame_bytes = view_bytes(frame)
        buffer = Buffer(self.renderer.device, len(frame_bytes), self.usage_flags, self.alloc_type, self.name)
        if buffer.is_mapped:
            buffer.data[:] = frame_bytes
        else:
            self.renderer.bulk_upload_list.append(BufferUploadInfo(frame_bytes, buffer, 0))
        return (buffer, GpuBufferView(buffer, 0, len(frame_bytes)))

    def _cmd_upload(self, cmd: CommandBuffer, staging_buffer: Buffer, view: GpuBufferView) -> None:
        cmd.copy_buffer_range(staging_buffer, view.buffer, view.size, 0, view.offset)

    def _append_barriers_and_mip_requests_for_upload(self, frame: RendererFrame, resource: V) -> None:
        frame.upload_property_pipeline_stages |= self.pipeline_stage_flags


class GpuImagePreuploadedProperty(GpuPreuploadedProperty[Image, GpuImageView]):
    def __init__(
        self,
        renderer: "Renderer",
        property: "ImageProperty",
        name: str,
    ):
        self.usage_flags = property.gpu_usage | ImageUsageFlags.TRANSFER_DST

        self.width = property.width
        self.height = property.height
        self.depth = property.depth if property.depth is not None else 1
        self.format = property.format
        self.layout = property.gpu_layout
        self.pipeline_stage_flags = property.gpu_stage
        self.srgb = property.gpu_srgb
        self.mips = property.gpu_mips

        # Created on promotion to dynamic, if needed
        self.spd_pipeline_instance: Optional[SPDPipelineInstance] = None

        self.pitch, self.rows, _ = get_image_pitch_rows_and_texel_size(self.width, self.height, self.format)

        super().__init__(
            self.depth * self.pitch * self.rows,
            False,  # batching is not supported for images
            renderer,
            renderer.image_upload_method,
            property,
            name,
        )

    # Private API
    def _promote_to_dynamic(self) -> None:
        if self.mips and self.spd_pipeline_instance is not None:
            self.spd_pipeline_instance = self.renderer.spd_pipeline.alloc_instance(self.renderer, False)

            mip_levels = max(self.width.bit_length(), self.height.bit_length())
            self.spd_pipeline_instance.set_image_extents(self.width, self.height, mip_levels)

        return super()._promote_to_dynamic()

    def _create_preuploaded_resources_and_views(
        self, frames: List[NDArray[Any]]
    ) -> Tuple[List[Image], List[GpuImageView]]:
        resources: List[Image] = []
        views: List[GpuImageView] = []

        for frame in frames:
            resource, view = self._create_preuploaded_resource_and_view_for_frame(frame)
            resources.append(resource)
            views.append(view)

        return resources, views

    def _create_preuploaded_resource_and_view_for_frame(self, frame: NDArray[Any]) -> Tuple[Image, GpuImageView]:
        view = _create_image_view(
            self.renderer.device,
            self.width,
            self.height,
            self.depth,
            self.format,
            self.usage_flags,
            self.srgb,
            self.mips,
            self.name,
        )
        img = view.image

        self.renderer.bulk_upload_list.append(
            ImageUploadInfo(
                view_bytes(frame),
                img,
                self.layout,
                view.mip_level_0_view,
                view.mip_views,
                MipGenerationFilter.AVERAGE_SRGB if self.srgb else MipGenerationFilter.AVERAGE,
            )
        )

        return img, view

    def _cmd_upload(self, cmd: CommandBuffer, staging_buffer: Buffer, view: GpuImageView) -> None:
        cmd.copy_buffer_to_image(staging_buffer, view.image)

    def _append_barriers_and_mip_requests_for_upload(self, frame: RendererFrame, resource: GpuImageView) -> None:
        frame.upload_before_image_barriers.append(
            ImageBarrier(
                resource.image,
                ImageLayout.UNDEFINED,
                ImageLayout.TRANSFER_DST_OPTIMAL,
                self.pipeline_stage_flags,
                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                PipelineStageFlags.COPY,
                AccessFlags.TRANSFER_WRITE,
            )
        )

        if self.mips:
            frame.upload_after_image_barriers.append(
                ImageBarrier(
                    resource.image,
                    ImageLayout.TRANSFER_DST_OPTIMAL,
                    ImageLayout.GENERAL,
                    PipelineStageFlags.COPY,
                    AccessFlags.TRANSFER_WRITE,
                    PipelineStageFlags.COMPUTE_SHADER,
                    AccessFlags.SHADER_SAMPLED_READ
                    | AccessFlags.SHADER_STORAGE_READ
                    | AccessFlags.SHADER_STORAGE_WRITE,
                )
            )
            mip_generation_filter = MipGenerationFilter.AVERAGE_SRGB if self.srgb else MipGenerationFilter.AVERAGE

            assert self.spd_pipeline_instance is not None
            assert resource.mip_level_0_view is not None

            descriptor_set = self.spd_pipeline_instance.get_and_write_current_and_advance(
                resource.mip_level_0_view, resource.mip_views
            )
            frame.mip_generation_requests.setdefault(mip_generation_filter, []).append(
                MipGenerationRequest(
                    resource.image,
                    self.spd_pipeline_instance.constants.tobytes(),
                    descriptor_set,
                    self.spd_pipeline_instance.groups_x,
                    self.spd_pipeline_instance.groups_y,
                    1,
                    ImageBarrier(
                        resource.image,
                        ImageLayout.GENERAL,
                        self.layout,
                        PipelineStageFlags.COMPUTE_SHADER,
                        AccessFlags.SHADER_SAMPLED_READ
                        | AccessFlags.SHADER_STORAGE_READ
                        | AccessFlags.SHADER_STORAGE_WRITE,
                        self.pipeline_stage_flags,
                        AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                    ),
                )
            )
        else:
            frame.upload_after_image_barriers.append(
                ImageBarrier(
                    resource.image,
                    ImageLayout.TRANSFER_DST_OPTIMAL,
                    self.layout,
                    PipelineStageFlags.COPY,
                    AccessFlags.TRANSFER_WRITE,
                    self.pipeline_stage_flags,
                    AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                )
            )


class GpuStreamingProperty(GpuProperty[V], Generic[R, V]):
    def __init__(
        self,
        renderer: "Renderer",
        upload_method: UploadMethod,
        property: "Property",
        pipeline_stage_flags: PipelineStageFlags,
        name: str,
    ):
        self.renderer = renderer
        self.upload_method = upload_method
        self.property = property
        self.pipeline_stage_flags = pipeline_stage_flags
        self.supported_operations = (
            GpuPropertySupportedOperations.UPDATE
            | GpuPropertySupportedOperations.APPEND
            | GpuPropertySupportedOperations.POP
            | GpuPropertySupportedOperations.INSERT
            | GpuPropertySupportedOperations.REMOVE
        )
        self.name = name

        # Variants
        self.async_load = self.property.upload.async_load
        self.mapped = (
            upload_method == UploadMethod.MAPPED_PREFER_DEVICE or upload_method == UploadMethod.MAPPED_PREFER_HOST
        )

        # Common
        self.current: Optional[V] = None

        # If not mapped
        self.current_gpu_res: Optional[GpuResource[V]] = None
        self.gpu_resources: List[GpuResource[V]] = []
        self.gpu_pool: Optional[LRUPool[int, GpuResource[V]]] = None

        # If prefetch and not mapped
        self.prefetch_states: List[PrefetchState] = []
        self.prefetch_states_lookup: Dict[GpuResource[V], PrefetchState] = {}

        # Compute buffer counts
        gpu_prefetch_count = (
            self.property.upload.gpu_prefetch_count if upload_method == UploadMethod.TRANSFER_QUEUE else 0
        )
        gpu_resources_count = 1 + gpu_prefetch_count
        cpu_prefetch_count = self.property.upload.cpu_prefetch_count
        cpu_buffers_count = self.renderer.num_frames_in_flight + cpu_prefetch_count + gpu_prefetch_count

        # Allocate CPU buffers
        self.current_cpu_buf: Optional[CpuBuffer] = None
        self.cpu_buffers = [
            CpuBuffer(self._create_cpu_buffer(f"{self.name}-cpu-{i}")) for i in range(cpu_buffers_count)
        ]

        self.cpu_pool: LRUPool[int, CpuBuffer] = LRUPool(
            self.cpu_buffers,
            self.renderer.num_frames_in_flight,
            cpu_prefetch_count,
        )

        if not self.mapped:
            for i in range(gpu_resources_count):
                res = self._create_gpu_resource(f"{name}-gpu-{i}")
                if upload_method == UploadMethod.TRANSFER_QUEUE:
                    semaphore = TimelineSemaphore(self.renderer.device, name=f"{name}-sem-{i}")
                else:
                    semaphore = None
                self.gpu_resources.append(GpuResource(res, semaphore))

            self.gpu_pool = LRUPool(
                self.gpu_resources,
                1,
                gpu_prefetch_count,
            )

            if upload_method == UploadMethod.TRANSFER_QUEUE:
                self.prefetch_states = [
                    PrefetchState(
                        commands=CommandBuffer(
                            self.renderer.device,
                            queue_family_index=self.renderer.device.transfer_queue_family_index,
                            name=f"{name}-prefetch-{i}",
                        ),
                        prefetch_done_value=0,
                    )
                    for i in range(gpu_prefetch_count)
                ]

    # Public API
    def get_current(self) -> V:
        assert self.current is not None
        return self.current

    def enqueue_for_destruction(self) -> None:
        self.renderer.enqueue_for_destruction(
            [cpu_buf.buf for cpu_buf in self.cpu_buffers] + [view.resource.resource() for view in self.gpu_resources]
        )

    # Renderer API
    def load(self, frame: RendererFrame) -> None:
        self.cpu_pool.release_frame(frame.index)

        def cpu_load(k: int, buf: CpuBuffer) -> None:
            if self.async_load:
                self.renderer.thread_pool.submit(buf.promise, self._load_async_into, k, buf)  # type: ignore
            else:
                buf.used_size = self.property.get_frame_by_index_into(k, buf.buf.data)

        property_frame_index = self.property.current_frame_index
        if self.mapped:
            self.current_cpu_buf = self.cpu_pool.get(property_frame_index, cpu_load)
        else:
            assert self.gpu_pool is not None

            def gpu_load(k: int, gpu_res: GpuResource[V]) -> None:
                assert self.current_cpu_buf is None

                cpu_buf = self.cpu_pool.get(property_frame_index, cpu_load)
                self.cpu_pool.use_frame(frame.index, k)

                self.current_cpu_buf = cpu_buf

                if self.upload_method == UploadMethod.GRAPHICS_QUEUE:
                    # Prepare for upload on graphics queue
                    self._append_barriers_for_upload_on_graphics_queue(frame, gpu_res.resource)
                    self._append_mip_generation_request(frame, gpu_res.resource)
                    gpu_res.state = GpuResourceState.RENDER
                else:
                    assert self.upload_method == UploadMethod.TRANSFER_QUEUE
                    assert (
                        gpu_res.state == GpuResourceState.EMPTY
                        or gpu_res.state == GpuResourceState.RENDER
                        or GpuResourceState.PREFETCH
                    ), gpu_res.state
                    assert frame.transfer_cmd is not None

                    frame.transfer_semaphores.append(gpu_res.use(PipelineStageFlags.TRANSFER))

                    # Prepare for upload on transfer queue
                    self._append_barriers_for_upload_on_transfer_queue(frame, gpu_res.resource)
                    self._append_release_barrier_on_transfer_queue(frame, gpu_res.resource)

                    gpu_res.state = GpuResourceState.LOAD

            def gpu_ensure(k: int, gpu_res: GpuResource[V]) -> None:
                assert gpu_res.state == GpuResourceState.PREFETCH, gpu_res.state

            gpu_res = self.gpu_pool.get(property_frame_index, gpu_load, gpu_ensure)
            self.gpu_pool.give_back(property_frame_index, gpu_res)

            if gpu_res.state == GpuResourceState.LOAD or gpu_res.state == GpuResourceState.PREFETCH:
                if gpu_res.state == GpuResourceState.PREFETCH:
                    assert frame.transfer_cmd is not None
                    frame.transfer_semaphores.append(gpu_res.use(PipelineStageFlags.ALL_COMMANDS))

                    # Sync: the resource was loaded by a prefetch operation on the transfer queue.
                    # Submit a release barrier on this frame to transfer ownership to this resource.
                    # The reason why this barrier needs to happen here instead of at the time of submission
                    # to the transfer queue, is because at that time we cannot guarantee that this resource
                    # will be used next on the graphics queue, the prefetch might be discarded and the resource
                    # used again later on the transfer queue. Since release-acquire barriers must be matched
                    # we issue the release only once we are sure that the acquire will happen.
                    self._append_release_barrier_on_transfer_queue(frame, gpu_res.resource)

                self._append_acquire_barrier_on_graphics_queue(frame, gpu_res.resource)
                self._append_mip_generation_request(frame, gpu_res.resource)
                gpu_res.state = GpuResourceState.RENDER

            assert gpu_res.state == GpuResourceState.RENDER, gpu_res.state

            # Sync: If we are using the transfer queue to upload we have to guard the gpu resource
            # with a semaphore because we might need to reuse this buffer for pre-fetching
            # while this frame is still in flight. Using the use_frame/release_frame mechanism
            # with a single frame is not enough because we might try to start pre-fetching on
            # the next frame while this frame is still in flight.
            if self.upload_method == UploadMethod.TRANSFER_QUEUE:
                frame.additional_semaphores.append(gpu_res.use(self.pipeline_stage_flags))

            self.current_gpu_res = gpu_res

    def upload(self, frame: RendererFrame) -> None:
        property_frame_index = self.property.current_frame_index

        # Wait for buffer to be ready if needed
        if self.async_load and self.current_cpu_buf is not None:
            self.current_cpu_buf.promise.get()

        # Wait for buffer to be ready. If this was already waited .get() is a no-op.
        if self.mapped:
            assert self.current_cpu_buf is not None
            cpu_buf = self.current_cpu_buf
            self.current_cpu_buf = None

            self.cpu_pool.use_frame(frame.index, property_frame_index)
            # NOTE: only works for buffers for now, which is why we get a type error here.
            self.current = GpuBufferView(cpu_buf.buf, 0, cpu_buf.used_size)  # type: ignore
        else:
            assert self.current_gpu_res is not None
            gpu_res = self.current_gpu_res
            self.current_gpu_res = None

            if self.current_cpu_buf is not None:
                cpu_buf = self.current_cpu_buf
                self.current_cpu_buf = None

                if self.upload_method == UploadMethod.GRAPHICS_QUEUE:
                    self._cmd_upload(frame.cmd, cpu_buf, gpu_res.resource)
                else:
                    assert self.upload_method == UploadMethod.TRANSFER_QUEUE
                    assert frame.transfer_cmd is not None
                    self._cmd_upload(frame.transfer_cmd, cpu_buf, gpu_res.resource)

            self.current = gpu_res.resource

    def prefetch(self) -> None:
        if self.property.upload.cpu_prefetch_count > 0:
            # Issue prefetches
            def cpu_prefetch_cleanup(k: int, buf: CpuBuffer) -> bool:
                return buf.promise.is_set()

            def cpu_prefetch(k: int, buf: CpuBuffer) -> None:
                self.renderer.thread_pool.submit(buf.promise, self._load_async_into, k, buf)  # type: ignore

            # TODO: can likely improve prefetch logic, and should probably allow
            # this to be hooked / configured somehow
            #
            # maybe we should just compute the state of next (maybe few) frames
            # (assuming constant dt, playback state) once globally and call prefetch with this.
            prefetch_start = self.property.current_frame_index + 1
            prefetch_end = prefetch_start + self.property.upload.cpu_prefetch_count
            prefetch_range = [
                self.property.animation.map_frame_index(i, 0, self.property.num_frames)
                for i in range(prefetch_start, prefetch_end)
            ]
            self.cpu_pool.prefetch(prefetch_range, cpu_prefetch_cleanup, cpu_prefetch)

        if self.upload_method == UploadMethod.TRANSFER_QUEUE and self.property.upload.gpu_prefetch_count > 0:
            assert self.gpu_pool is not None

            def gpu_prefetch_cleanup(k: int, gpu_res: GpuResource[V]) -> bool:
                state = self.prefetch_states_lookup[gpu_res]

                assert gpu_res.semaphore is not None

                if gpu_res.semaphore.get_value() >= state.prefetch_done_value:
                    # Release prefetch state
                    self.prefetch_states.append(state)
                    del self.prefetch_states_lookup[gpu_res]

                    # Release buffer
                    self.cpu_pool.release_manual(k)

                    assert gpu_res.state == GpuResourceState.RENDER or gpu_res.state == GpuResourceState.PREFETCH
                    return True
                return False

            def gpu_prefetch(k: int, gpu_res: GpuResource[V]) -> None:
                assert (
                    gpu_res.state == GpuResourceState.EMPTY
                    or GpuResourceState.PREFETCH
                    or gpu_res.state == GpuResourceState.RENDER
                ), gpu_res.state

                # We know that the cpu buffer is available, so just get it
                cpu_next: CpuBuffer = self.cpu_pool.get(k, lambda _x, _y: None)
                self.cpu_pool.use_manual(k)

                # Get free prefetch state
                state = self.prefetch_states.pop()
                self.prefetch_states_lookup[gpu_res] = state

                with state.commands:
                    # Sync: Flush previous upload on gfx queue and transition images.
                    self._cmd_before_barrier(state.commands, gpu_res.resource)
                    self._cmd_upload(state.commands, cpu_next, gpu_res.resource)

                info = gpu_res.use(PipelineStageFlags.TRANSFER)

                # Prefetch commands are submitted one at a time because we have no
                # knowledge or how far in the future they will be used. If they were
                # batched you would need to wait for all of them to complete before
                # you can use any of the resources. These submissions are potentially
                # expensive, but we assume that applications will make very sparse
                # use of prefetching, only for really large resources of which hopeufully
                # there are only one or a few.
                self.renderer.device.transfer_queue.submit(
                    state.commands,
                    wait_semaphores=[(info.sem, info.wait_value, info.wait_stage)],
                    signal_semaphores=[(info.sem, info.signal_value, info.signal_stage)],
                )
                state.prefetch_done_value = info.signal_value
                gpu_res.state = GpuResourceState.PREFETCH

            # TODO: fix, same as above
            prefetch_start = self.property.current_frame_index + 1
            prefetch_end = prefetch_start + self.property.upload.gpu_prefetch_count
            available_range = []
            for i in range(prefetch_start, prefetch_end):
                frame_index = self.property.animation.map_frame_index(i, 0, self.property.num_frames)
                if self.cpu_pool.is_available(frame_index):
                    available_range.append(frame_index)

            self.gpu_pool.prefetch(
                available_range,
                gpu_prefetch_cleanup,
                gpu_prefetch,
            )

    # Edit API
    def update_frame(self, index: int) -> None:
        self._invalidate_frame(index)

    def update_frames(self, indices: List[int]) -> None:
        self._invalidate_all_frames()

    def update_frame_range(self, start: int, stop: int) -> None:
        self._invalidate_all_frames()

    def append_frame(self) -> None:
        # Nothing to do here
        pass

    def append_frames(self, count: int) -> None:
        # Nothing to do here
        pass

    def pop_frame(self) -> None:
        self._invalidate_frame(self.property.num_frames - 1)

    def pop_frames(self, count: int) -> None:
        self._invalidate_all_frames()

    def insert_frame(self, index: int) -> None:
        self._invalidate_all_frames()

    def insert_frames(self, index: int, count: int) -> None:
        self._invalidate_all_frames()

    def remove_frame(self, index: int) -> None:
        self._invalidate_all_frames()

    def remove_frames(self, indices: List[int]) -> None:
        self._invalidate_all_frames()

    def remove_frame_range(self, start: int, stop: int) -> None:
        self._invalidate_all_frames()

    # Private API
    def _load_async(self, i: int, thread_index: int) -> "PropertyItem":
        return self.property.get_frame_by_index(i, thread_index)

    def _load_async_into(self, i: int, buf: CpuBuffer, thread_index: int) -> None:
        buf.used_size = self.property.get_frame_by_index_into(i, buf.buf.data, thread_index)

    def _invalidate_frame(self, index: int) -> None:
        self.cpu_pool.increment_generation(index)
        if self.gpu_pool is not None:
            self.gpu_pool.increment_generation(index)

    def _invalidate_all_frames(self) -> None:
        self.cpu_pool.increment_all_generations()
        if self.gpu_pool is not None:
            self.gpu_pool.increment_all_generations()

    def _create_cpu_buffer(self, name: str) -> Buffer:
        raise NotImplementedError

    def _create_gpu_resource(self, name: str) -> V:
        raise NotImplementedError

    def _append_mip_generation_request(self, frame: RendererFrame, resource: V) -> None:
        pass

    def _append_barriers_for_upload_on_graphics_queue(self, frame: RendererFrame, resource: V) -> None:
        raise NotImplementedError

    def _append_barriers_for_upload_on_transfer_queue(self, frame: RendererFrame, resource: V) -> None:
        raise NotImplementedError

    def _append_acquire_barrier_on_graphics_queue(self, frame: RendererFrame, resource: V) -> None:
        raise NotImplementedError

    def _append_release_barrier_on_transfer_queue(self, frame: RendererFrame, resource: V) -> None:
        raise NotImplementedError

    def _cmd_upload(self, cmd: CommandBuffer, cpu_buf: CpuBuffer, resource: V) -> None:
        raise NotImplementedError

    def _cmd_before_barrier(self, cmd: CommandBuffer, resource: V) -> None:
        pass


class GpuBufferStreamingProperty(GpuStreamingProperty[Buffer, GpuBufferView]):
    def __init__(
        self,
        renderer: "Renderer",
        property: "BufferProperty",
        name: str,
    ):
        self.usage_flags = property.gpu_usage | BufferUsageFlags.TRANSFER_DST
        self.max_frame_size = property.max_size

        super().__init__(
            renderer,
            renderer.buffer_upload_method,
            property,
            property.gpu_stage,
            name,
        )

    # Private API
    def _create_cpu_buffer(self, name: str) -> Buffer:
        cpu_alloc_type = (
            AllocType.DEVICE_MAPPED if self.upload_method == UploadMethod.MAPPED_PREFER_DEVICE else AllocType.HOST
        )
        if self.mapped:
            cpu_buffer_usage_flags = self.usage_flags
        else:
            cpu_buffer_usage_flags = BufferUsageFlags.TRANSFER_SRC
        return Buffer(self.renderer.device, self.max_frame_size, cpu_buffer_usage_flags, cpu_alloc_type, name)

    def _create_gpu_resource(self, name: str) -> GpuBufferView:
        return GpuBufferView(
            Buffer(
                self.renderer.device,
                self.max_frame_size,
                self.usage_flags,
                AllocType.DEVICE,
                name=name,
            ),
            0,
            self.max_frame_size,
        )

    def _append_barriers_for_upload_on_graphics_queue(self, frame: RendererFrame, resource: GpuBufferView) -> None:
        frame.upload_property_pipeline_stages |= self.pipeline_stage_flags

    def _append_barriers_for_upload_on_transfer_queue(self, frame: RendererFrame, resource: GpuBufferView) -> None:
        pass

    def _append_acquire_barrier_on_graphics_queue(self, frame: RendererFrame, resource: GpuBufferView) -> None:
        frame.upload_after_buffer_barriers.append(
            BufferBarrier(
                resource.buffer,
                PipelineStageFlags.NONE,
                AccessFlags.NONE,
                self.pipeline_stage_flags,
                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                self.renderer.device.transfer_queue_family_index,
                self.renderer.device.graphics_queue_family_index,
            )
        )

    def _append_release_barrier_on_transfer_queue(self, frame: RendererFrame, resource: GpuBufferView) -> None:
        frame.transfer_upload_after_buffer_barriers.append(
            BufferBarrier(
                resource.buffer,
                PipelineStageFlags.COPY,
                AccessFlags.TRANSFER_WRITE,
                PipelineStageFlags.NONE,
                AccessFlags.NONE,
                self.renderer.device.transfer_queue_family_index,
                self.renderer.device.graphics_queue_family_index,
            )
        )

    def _cmd_upload(
        self,
        cmd: CommandBuffer,
        cpu_buf: CpuBuffer,
        resource: GpuBufferView,
    ) -> None:
        cmd.copy_buffer_range(cpu_buf.buf, resource.buffer, cpu_buf.used_size, resource.offset)
        resource.size = cpu_buf.used_size

    def _cmd_before_barrier(self, cmd: CommandBuffer, resource: GpuBufferView) -> None:
        # Needs to be all when going from application usage on graphics queue to copy on transfer queue for prefetching
        cmd.memory_barrier(MemoryUsage.ALL, MemoryUsage.TRANSFER_DST)


class GpuImageStreamingProperty(GpuStreamingProperty[Image, GpuImageView]):
    def __init__(
        self,
        renderer: "Renderer",
        property: "ImageProperty",
        name: str,
    ):
        self.usage_flags = property.gpu_usage | ImageUsageFlags.TRANSFER_DST

        self.width = property.width
        self.height = property.height
        self.depth = property.depth if property.depth is not None else 1
        self.format = property.format
        self.layout = property.gpu_layout
        self.srgb = property.gpu_srgb
        self.mips = property.gpu_mips

        self.spd_pipeline_instance = None
        if self.mips:
            assert self.depth == 1
            mip_levels = max(self.width.bit_length(), self.height.bit_length())
            self.spd_pipeline_instance = renderer.spd_pipeline.alloc_instance(renderer, False) if self.mips else None
            self.spd_pipeline_instance.set_image_extents(self.width, self.height, mip_levels)

        self.pitch, self.rows, _ = get_image_pitch_rows_and_texel_size(self.width, self.height, self.format)

        super().__init__(
            renderer,
            renderer.image_upload_method,
            property,
            property.gpu_stage,
            name,
        )

    # Private API
    def _create_cpu_buffer(self, name: str) -> Buffer:
        return Buffer(
            self.renderer.device,
            self.depth * self.pitch * self.rows,
            BufferUsageFlags.TRANSFER_SRC,
            AllocType.HOST,
            name,
        )

    def _create_gpu_resource(self, name: str) -> GpuImageView:
        return _create_image_view(
            self.renderer.device,
            self.width,
            self.height,
            self.depth,
            self.format,
            self.usage_flags,
            self.srgb,
            self.mips,
            name,
        )

    def _append_mip_generation_request(self, frame: RendererFrame, resource: GpuImageView) -> None:
        if not self.mips:
            return

        mip_generation_filter = MipGenerationFilter.AVERAGE_SRGB if self.srgb else MipGenerationFilter.AVERAGE

        assert self.spd_pipeline_instance is not None
        assert resource.mip_level_0_view is not None

        descriptor_set = self.spd_pipeline_instance.get_and_write_current_and_advance(
            resource.mip_level_0_view, resource.mip_views
        )
        frame.mip_generation_requests.setdefault(mip_generation_filter, []).append(
            MipGenerationRequest(
                resource.image,
                self.spd_pipeline_instance.constants.tobytes(),
                descriptor_set,
                self.spd_pipeline_instance.groups_x,
                self.spd_pipeline_instance.groups_y,
                1,
                ImageBarrier(
                    resource.image,
                    ImageLayout.GENERAL,
                    self.layout,
                    PipelineStageFlags.COMPUTE_SHADER,
                    AccessFlags.SHADER_SAMPLED_READ
                    | AccessFlags.SHADER_STORAGE_READ
                    | AccessFlags.SHADER_STORAGE_WRITE,
                    self.pipeline_stage_flags,
                    AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                ),
            )
        )

    def _append_barriers_for_upload_on_graphics_queue(self, frame: RendererFrame, resource: GpuImageView) -> None:
        frame.upload_before_image_barriers.append(
            ImageBarrier(
                resource.image,
                ImageLayout.UNDEFINED,
                ImageLayout.TRANSFER_DST_OPTIMAL,
                self.pipeline_stage_flags,
                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                PipelineStageFlags.COPY,
                AccessFlags.TRANSFER_WRITE,
            )
        )

        if self.mips:
            frame.upload_after_image_barriers.append(
                ImageBarrier(
                    resource.image,
                    ImageLayout.TRANSFER_DST_OPTIMAL,
                    ImageLayout.GENERAL,
                    PipelineStageFlags.COPY,
                    AccessFlags.TRANSFER_WRITE,
                    PipelineStageFlags.COMPUTE_SHADER,
                    AccessFlags.SHADER_SAMPLED_READ
                    | AccessFlags.SHADER_STORAGE_READ
                    | AccessFlags.SHADER_STORAGE_WRITE,
                )
            )
        else:
            frame.upload_after_image_barriers.append(
                ImageBarrier(
                    resource.image,
                    ImageLayout.TRANSFER_DST_OPTIMAL,
                    self.layout,
                    PipelineStageFlags.COPY,
                    AccessFlags.TRANSFER_WRITE,
                    self.pipeline_stage_flags,
                    AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                )
            )

    def _append_barriers_for_upload_on_transfer_queue(self, frame: RendererFrame, resource: GpuImageView) -> None:
        frame.transfer_upload_before_image_barriers.append(
            ImageBarrier(
                resource.image,
                ImageLayout.UNDEFINED,
                ImageLayout.TRANSFER_DST_OPTIMAL,
                # Needs to be all when going from application usage on graphics queue to copy on transfer queue.
                PipelineStageFlags.ALL_COMMANDS,
                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                PipelineStageFlags.COPY,
                AccessFlags.TRANSFER_WRITE,
            )
        )

    def _append_acquire_barrier_on_graphics_queue(self, frame: RendererFrame, resource: GpuImageView) -> None:
        if self.mips:
            frame.upload_after_image_barriers.append(
                ImageBarrier(
                    resource.image,
                    ImageLayout.TRANSFER_DST_OPTIMAL,
                    ImageLayout.GENERAL,
                    PipelineStageFlags.COPY,
                    AccessFlags.TRANSFER_WRITE,
                    PipelineStageFlags.COMPUTE_SHADER,
                    AccessFlags.SHADER_SAMPLED_READ
                    | AccessFlags.SHADER_STORAGE_READ
                    | AccessFlags.SHADER_STORAGE_WRITE,
                    self.renderer.device.transfer_queue_family_index,
                    self.renderer.device.graphics_queue_family_index,
                )
            )
        else:
            frame.upload_after_image_barriers.append(
                ImageBarrier(
                    resource.image,
                    # Layout transition should be specified twice the same for acquire/release pairs
                    # according to the vulkan synchronization examples.
                    ImageLayout.TRANSFER_DST_OPTIMAL,
                    self.layout,
                    # The spec says that these should be ignored, but validation seems to
                    # look at these when a layout transition is involved.
                    PipelineStageFlags.COPY,
                    AccessFlags.TRANSFER_WRITE,
                    self.pipeline_stage_flags,
                    AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                    self.renderer.device.transfer_queue_family_index,
                    self.renderer.device.graphics_queue_family_index,
                )
            )

    def _append_release_barrier_on_transfer_queue(self, frame: RendererFrame, resource: GpuImageView) -> None:
        frame.transfer_upload_after_image_barriers.append(
            ImageBarrier(
                resource.image,
                # Layout transition should be specified twice the same for acquire/release pairs
                # according to the vulkan synchronization examples.
                ImageLayout.TRANSFER_DST_OPTIMAL,
                ImageLayout.GENERAL if self.mips else self.layout,
                PipelineStageFlags.TRANSFER,
                AccessFlags.TRANSFER_WRITE,
                # The spec says that these should be ignored, but validation seems to
                # look at these when a layout transition is involved.
                PipelineStageFlags.COPY,
                AccessFlags.TRANSFER_WRITE,
                self.renderer.device.transfer_queue_family_index,
                self.renderer.device.graphics_queue_family_index,
            )
        )

    def _cmd_upload(self, cmd: CommandBuffer, buffer: CpuBuffer, view: GpuImageView) -> None:
        cmd.copy_buffer_to_image(buffer.buf, view.image)

    def _cmd_before_barrier(self, cmd: CommandBuffer, resource: GpuImageView) -> None:
        cmd.image_barrier(
            resource.image,
            ImageLayout.TRANSFER_DST_OPTIMAL,
            # Needs to be all when going from application usage on graphics queue to copy on transfer queue for prefetching
            MemoryUsage.ALL,
            MemoryUsage.TRANSFER_DST,
            undefined=True,
        )



from typing import Optional

from pyglm.glm import clamp, floor, ivec2, vec2
from pyxpg import (
    BorderColor,
    DescriptorSetBinding,
    DescriptorType,
    ImageLayout,
    ImageUsageFlags,
    PipelineStageFlags,
    Sampler,
    SamplerAddressMode,
    Stage,
    imgui,
)

from ambra.renderer_frame import RendererFrame
from ambra.utils.ring_buffer import RingBuffer

from .property import ImageProperty
from .renderer import Renderer
from .scene import Widget
from .utils.descriptors import create_descriptor_layout_pool_and_sets


class ImageInspector(Widget):
    def __init__(self, title: str, image: ImageProperty):
        super().__init__(title)

        self.image = self.add_image_property(image, name="image").use_gpu(
            ImageUsageFlags.SAMPLED,
            ImageLayout.SHADER_READ_ONLY_OPTIMAL,
            PipelineStageFlags.FRAGMENT_SHADER,
            mips=True,
        )

        self.zoom = 0.0
        self.zoom_speed = 0.25
        self.offset = vec2(0, 0)
        self.dragging = False
        self.drag_start_pos = ivec2(0, 0)
        self.drag_start_offset = vec2(0, 0)

        # Public API
        self.pixel_under_cursor: Optional[ivec2] = None

    def create(self, r: Renderer) -> None:
        self.sampler = Sampler(
            r.device,
            u=SamplerAddressMode.CLAMP_TO_BORDER,
            v=SamplerAddressMode.CLAMP_TO_BORDER,
            border_color=BorderColor.FLOAT_OPAQUE_BLACK,
        )

        self.descriptor_layout, self.descriptor_pool, self.descriptor_sets = create_descriptor_layout_pool_and_sets(
            r.device,
            [
                DescriptorSetBinding(1, DescriptorType.COMBINED_IMAGE_SAMPLER, stage_flags=Stage.FRAGMENT),
            ],
            r.num_frames_in_flight,
        )
        self.texture_and_sets = RingBuffer([(s, imgui.Texture(s)) for s in self.descriptor_sets])

    def upload(self, r: Renderer, frame: RendererFrame) -> None:
        s, self.texture = self.texture_and_sets.get_current_and_advance()
        s.write_combined_image_sampler(
            self.image.get_current_gpu().image, ImageLayout.SHADER_READ_ONLY_OPTIMAL, self.sampler, 0
        )

    def _draw(self) -> None:
        self.pixel_under_cursor = None

        image_size = ivec2(self.image.width, self.image.height)
        image_visible_size = vec2(image_size) * float(2.0**self.zoom)

        ar = image_visible_size.x / image_visible_size.y
        avail = imgui.get_content_region_avail()
        if avail.x <= 0 or avail.y <= 0:
            return

        available = ivec2(avail.x, avail.y)
        height = min(available.x / ar, available.y)
        view_size = ivec2(ar * height, height)

        io = imgui.get_io()

        mouse_pos = ivec2(io.mouse_pos.x, io.mouse_pos.y)
        pixel_size = vec2(view_size) / image_visible_size
        if self.dragging:
            self.offset = self.drag_start_offset - vec2(mouse_pos - self.drag_start_pos) / pixel_size

        draw_list = imgui.get_window_draw_list()

        cursor_pos = imgui.get_cursor_screen_pos()
        pos = ivec2(cursor_pos.x, cursor_pos.y)
        mouse_relative_pos = mouse_pos - pos

        image_top_left = vec2(self.offset)
        image_bottom_right = image_top_left + image_visible_size

        imgui.image(
            self.texture,
            imgui.Vec2(*view_size),
            imgui.Vec2(*(image_top_left / vec2(image_size))),
            imgui.Vec2(*(image_bottom_right / vec2(image_size))),
        )

        if imgui.is_item_hovered() and io.mouse_wheel != 0:
            self.zoom -= io.mouse_wheel * self.zoom_speed

            before_mouse_pixel_coordinates = vec2(mouse_relative_pos) / pixel_size
            after_pixel_size = vec2(view_size) / (vec2(image_size) * (2.0**self.zoom))
            after_mouse_pixel_coordinates = vec2(mouse_relative_pos) / after_pixel_size

            self.offset -= after_mouse_pixel_coordinates - before_mouse_pixel_coordinates

        if imgui.is_item_clicked(imgui.MouseButton.LEFT):
            self.dragging = True
            self.drag_start_pos = mouse_pos
            self.drag_start_offset = vec2(self.offset)
        elif not imgui.is_mouse_down(imgui.MouseButton.LEFT):
            self.dragging = False

        if (
            mouse_relative_pos.x >= 0
            and mouse_relative_pos.y >= 0
            and mouse_relative_pos.x < view_size.x
            and mouse_relative_pos.y < view_size.y
        ):
            window_rect_pos = vec2(pos)
            window_rect_end_pos = vec2(pos) + vec2(view_size)

            pixel_coordinates = vec2(mouse_relative_pos) / pixel_size + image_top_left
            pixel_pos = window_rect_pos + (vec2(ivec2(pixel_coordinates)) - image_top_left) * pixel_size
            pixel_image_coordinates = ivec2(floor(pixel_coordinates))

            if (
                pixel_image_coordinates.x >= 0
                and pixel_image_coordinates.y >= 0
                and pixel_image_coordinates.x < image_size.x
                and pixel_image_coordinates.y < image_size.y
            ):
                pixel_rect_pos = pixel_pos
                pixel_rect_end_pos = pixel_pos + pixel_size

                pixel_rect_pos = clamp(pixel_rect_pos, window_rect_pos, window_rect_end_pos)
                pixel_rect_end_pos = clamp(pixel_rect_end_pos, window_rect_pos, window_rect_end_pos)

                draw_list.add_rect(
                    imgui.Vec2(*pixel_rect_pos), imgui.Vec2(*pixel_rect_end_pos), 0xFF00FFFF, thickness=2
                )

                imgui.set_next_window_pos(imgui.Vec2(pixel_rect_end_pos.x + 1.0, pixel_rect_end_pos.y + 1.0))
                imgui.begin_tooltip()
                self.tooltip(pixel_image_coordinates)
                imgui.end_tooltip()

                self.pixel_under_cursor = pixel_image_coordinates

        draw_list.add_rect(cursor_pos, imgui.Vec2(*(pos + view_size)), 0xFFFFFFFF, thickness=2)

        imgui.set_cursor_screen_pos(cursor_pos)
        imgui.invisible_button("###invisible", imgui.Vec2(*view_size))

    def tooltip(self, pixel_image_coordinates: ivec2) -> None:
        img_data = self.image.get_current()
        values = img_data[pixel_image_coordinates.y, pixel_image_coordinates.x]
        values_text = ", ".join([str(v) for v in values])
        imgui.text(f"({pixel_image_coordinates.x}, {pixel_image_coordinates.y}): [{values_text}]")

    def gui(self) -> None:
        if imgui.begin(f"{self.title}###widget-{self.uid}")[0]:  # noqa: SIM102
            # NOTE: widgets are not disabled if their properties are not enabled to avoid UI appearing and disappearing.
            # They are supposed to handle missing properties gracefully in their implementation.
            if self.image.current_animation_enabled:
                self._draw()
        imgui.end()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from pyxpg import Key, Modifiers, MouseButton


@dataclass(frozen=True)
class KeyBinding:
    key: Key
    mods: Modifiers = Modifiers.NONE

    def is_active(self, key: Key, mods: Modifiers) -> bool:
        return self.key == key and self.mods == mods


@dataclass(frozen=True)
class MouseButtonBinding:
    button: MouseButton
    mods: Modifiers = Modifiers.NONE

    def is_active(self, button: MouseButton, mods: Modifiers) -> bool:
        return self.button == button and self.mods == mods


@dataclass
class KeyMap:
    exit: KeyBinding = KeyBinding(Key.ESCAPE)

    # Playback
    toggle_play_pause: KeyBinding = KeyBinding(Key.SPACE)
    next_frame: KeyBinding = KeyBinding(Key.PERIOD)
    previous_frame: KeyBinding = KeyBinding(Key.COMMA)

    # Renderer
    toggle_path_tracer: KeyBinding = KeyBinding(Key.R, Modifiers.CTRL)

    # Camera control
    camera_rotate: MouseButtonBinding = MouseButtonBinding(MouseButton.LEFT)
    camera_pan: MouseButtonBinding = MouseButtonBinding(MouseButton.RIGHT)
    camera_zoom_modifiers: Modifiers = Modifiers.NONE
    camera_zoom_move_modifiers: Modifiers = Modifiers.SHIFT



import numpy as np
from pyglm.glm import vec3

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig
from ambra.geometry import create_sphere
from ambra.materials import ColorMaterial
from ambra.primitives3d import Mesh
from ambra.property import (
    AnimationBoundary,
    ArrayBufferProperty,
    FrameSampledAnimation,
    TimeSampledAnimation,
)
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
            max_time=3.0,
        ),
        camera=CameraConfig(
            position=vec3(0, 5, 0),
            target=vec3(0),
        ),
        gui=GuiConfig(
            stats=True,
            playback=True,
            inspector=True,
        ),
    ),
)

v, _, f = create_sphere(0.5, rings=32, sectors=64)

enabled_always = True

# Disabled at frame 0, enabled at frame 20, disabled again at frame 40
enabled_frame_range = ArrayBufferProperty(
    [1, 0],
    bool,
    animation=FrameSampledAnimation(
        [20, 40],
        AnimationBoundary.HOLD,
    ),
)

# Disabled at time 0s, enabled at time 1s, disabled again at time 2s
enabled_timestamp_range = ArrayBufferProperty(
    [1, 0],
    bool,
    animation=TimeSampledAnimation([1.0, 2.0], AnimationBoundary.HOLD),
)

always = Mesh(v, f, translation=(1.5, 0, 0.7), material=ColorMaterial((0.8, 0, 0)), enabled=enabled_always)
frame_range = Mesh(v, f, translation=(0, 0, 0.7), material=ColorMaterial((0, 0.8, 0)), enabled=enabled_frame_range)
timestamp_range = Mesh(
    v, f, translation=(-1.5, 0, 0.7), material=ColorMaterial((0, 0, 0.8)), enabled=enabled_timestamp_range
)

v_enabled_frame_range = ArrayBufferProperty(
    [v], np.float32, animation=FrameSampledAnimation([20, 40], boundary=AnimationBoundary.DISABLE)
)
v_enabled_timestamp_range = ArrayBufferProperty(
    [v], np.float32, animation=TimeSampledAnimation([1.0, 2.0], boundary=AnimationBoundary.DISABLE)
)

new_always = Mesh(v, f, translation=(1.5, 0, -0.7), material=ColorMaterial((0.8, 0, 0)))
new_frame_range = Mesh(v_enabled_frame_range, f, translation=(0, 0, -0.7), material=ColorMaterial((0, 0.8, 0)))
new_timestamp_range = Mesh(
    v_enabled_timestamp_range, f, translation=(-1.5, 0, -0.7), material=ColorMaterial((0, 0, 0.8))
)


viewer.scene.objects.extend(
    [
        always,
        frame_range,
        timestamp_range,
        new_always,
        new_frame_range,
        new_timestamp_range,
    ]
)

viewer.run()



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Union

import numpy as np
from numpy.typing import NDArray
from pyxpg import (
    AllocType,
    Buffer,
    BufferUsageFlags,
    CommandBuffer,
    ComputePipeline,
    DescriptorPool,
    DescriptorPoolCreateFlags,
    DescriptorSet,
    DescriptorSetBinding,
    DescriptorSetLayout,
    DescriptorSetLayoutCreateFlags,
    DescriptorType,
    DeviceFeatures,
    Filter,
    Format,
    Image,
    ImageBarrier,
    ImageLayout,
    ImageUsageFlags,
    ImageView,
    ImageViewType,
    MemoryUsage,
    PushConstantsRange,
    Sampler,
    SamplerAddressMode,
    Shader,
    SubgroupFeatureFlags,
)

from .utils.descriptors import create_descriptor_pool_and_sets_ringbuffer
from .utils.gpu import MipGenerationFilter, view_bytes
from .utils.ring_buffer import RingBuffer

if TYPE_CHECKING:
    from .renderer import Renderer

SPD_MAX_LEVELS = 12


class SPDPipelineInstance:
    def __init__(
        self,
        atomic_counters: Buffer,
        descriptor_pool: DescriptorPool,
        descriptor_sets: RingBuffer[DescriptorSet],
        constants: NDArray[Any],
    ):
        self.atomic_counters = atomic_counters
        self.descriptor_pool = descriptor_pool
        self.descriptor_sets = descriptor_sets
        self.constants = constants

        self.groups_x = 0
        self.groups_y = 0

    def set_image_extents(self, width: int, height: int, mips: int, x: int = 0, y: int = 0) -> None:
        rect_width = width
        rect_height = height

        group_offset_x = x // 64
        group_offset_y = y // 64

        group_end_x = (x + rect_width - 1) // 64
        group_end_y = (y + rect_height - 1) // 64

        groups_x = group_end_x + 1 - group_offset_x
        groups_y = group_end_y + 1 - group_offset_y

        groups = groups_x * groups_y

        # image.mip_levels includes level 0. The kernel expects just the number of levels after 0.
        mips_after_level_0 = min(mips - 1, SPD_MAX_LEVELS)

        self.constants["mips"] = mips_after_level_0
        self.constants["numWorkGroups"] = groups
        self.constants["workGroupOffset"] = (group_offset_x, group_offset_y)
        self.constants["invInputSize"] = (1.0 / width, 1.0 / height)

        self.groups_x = groups_x
        self.groups_y = groups_y

    def get_and_write_current_and_advance(
        self, level_0_view: Union[Image, ImageView], mip_views: List[ImageView]
    ) -> DescriptorSet:
        s = self.descriptor_sets.get_current_and_advance()
        s.write_image(level_0_view, ImageLayout.GENERAL, DescriptorType.SAMPLED_IMAGE, 2)
        for m in range(SPD_MAX_LEVELS + 1):
            view = mip_views[m] if m < len(mip_views) else mip_views[0]
            if m == 6:
                s.write_image(view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 3, 0)
            s.write_image(view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 4, m)
        return s


@dataclass
class MipGenerationRequest:
    image: Image
    push_constants: bytes
    descriptor_set: DescriptorSet
    groups_x: int
    groups_y: int
    groups_z: int
    after_barrier: ImageBarrier


class SPDPipeline:
    def __init__(self, r: "Renderer"):
        from .renderer import SHADERS_PATH  # noqa: PLC0415

        self.constants_dtype = np.dtype(
            {
                "mips": (np.uint32, 0),
                "numWorkGroups": (np.uint32, 4),
                "workGroupOffset": (np.dtype((np.uint32, 2)), 8),
                "invInputSize": (np.dtype((np.float32, 2)), 16),
            }
        )  # type: ignore

        # NOTE: using UPDATE_AFTER_BIND because that allows for a larger limit on total
        # number of bound storage images in MoltenVK. This is because MoltenVK switches
        # to using argument buffers when this flag is set. Not sure why it does not
        # expose the higher limit anyways when argument buffers are available.
        # Newer versions of MoltenVK seem to have increased this limit anyway.
        #
        # See: https://github.com/KhronosGroup/MoltenVK/issues/1610
        self.descriptor_set_layout = DescriptorSetLayout(
            r.device,
            [
                DescriptorSetBinding(1, DescriptorType.SAMPLER),
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),
                DescriptorSetBinding(1, DescriptorType.SAMPLED_IMAGE),
                DescriptorSetBinding(1, DescriptorType.STORAGE_IMAGE),
                DescriptorSetBinding(SPD_MAX_LEVELS + 1, DescriptorType.STORAGE_IMAGE),
            ],
            DescriptorSetLayoutCreateFlags.UPDATE_AFTER_BIND_POOL,
        )

        # Check hardware support for float16 and quad operations. Otherwise fallback to LDS.
        use_float16 = (r.device.features & DeviceFeatures.SHADER_FLOAT16_INT8) != 0
        use_wave_ops = (
            r.device.device_properties.subgroup_properties.supported_operations & SubgroupFeatureFlags.QUAD
        ) != 0 and (r.device.features & DeviceFeatures.SHADER_SUBGROUP_EXTENDED_TYPES) != 0

        defines = [
            ("FFX_GPU", ""),
            ("FFX_SLANG", ""),
            ("FFX_HLSL_SM", "62"),
            ("FFX_SPD_OPTION_LINEAR_SAMPLE", "1"),
            ("FFX_HALF", "1" if use_float16 else "0"),
            ("FFX_NO_16_BIT_CAST", ""),
            ("FFX_SPD_OPTION_WAVE_INTEROP_LDS", "0" if use_wave_ops else "1"),
        ]
        self.avg_shader = r.compile_builtin_shader(
            "ffx/ffx_spd_downsample_pass.slang",
            entry="CS",
            defines=[*defines, ("FFX_SPD_OPTION_DOWNSAMPLE_FILTER", "0")],
            include_paths=[SHADERS_PATH.joinpath("ffx")],
        )
        self.avg_srgb_shader = r.compile_builtin_shader(
            "ffx/ffx_spd_downsample_pass.slang",
            entry="CS",
            defines=[*defines, ("FFX_SPD_OPTION_DOWNSAMPLE_FILTER", "0"), ("FFX_SPD_SRGB", "1")],
            include_paths=[SHADERS_PATH.joinpath("ffx")],
        )
        self.min_shader = r.compile_builtin_shader(
            "ffx/ffx_spd_downsample_pass.slang",
            entry="CS",
            defines=[*defines, ("FFX_SPD_OPTION_DOWNSAMPLE_FILTER", "1")],
            include_paths=[SHADERS_PATH.joinpath("ffx")],
        )
        self.max_shader = r.compile_builtin_shader(
            "ffx/ffx_spd_downsample_pass.slang",
            entry="CS",
            defines=[*defines, ("FFX_SPD_OPTION_DOWNSAMPLE_FILTER", "2")],
            include_paths=[SHADERS_PATH.joinpath("ffx")],
        )

        self.avg_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.avg_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )
        self.avg_srgb_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.avg_srgb_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )
        self.min_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.min_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )
        self.max_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.max_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

        self.linear_clamp_sampler = Sampler(
            r.device,
            min_filter=Filter.LINEAR,
            mag_filter=Filter.LINEAR,
            u=SamplerAddressMode.CLAMP_TO_EDGE,
            v=SamplerAddressMode.CLAMP_TO_EDGE,
            w=SamplerAddressMode.CLAMP_TO_EDGE,
        )

        self.sync_instance = self.alloc_instance(r, single_set=True)

    def alloc_instance(self, r: "Renderer", single_set: bool = False) -> SPDPipelineInstance:
        constants = np.zeros((1,), self.constants_dtype)

        atomic_counters = Buffer.from_data(
            r.device,
            view_bytes(np.zeros((6,), np.uint32)),
            BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            name="ffx-spd-atomic-counters",
        )

        number_of_sets = r.num_frames_in_flight if not single_set else 1
        pool, descriptor_sets = create_descriptor_pool_and_sets_ringbuffer(
            r.device,
            self.descriptor_set_layout,
            number_of_sets,
            pool_flags=DescriptorPoolCreateFlags.UPDATE_AFTER_BIND,
            name="spd-instance",
        )

        for s in descriptor_sets:
            s.write_sampler(self.linear_clamp_sampler, 0)
            s.write_buffer(atomic_counters, DescriptorType.STORAGE_BUFFER, 1)

        return SPDPipelineInstance(atomic_counters, pool, descriptor_sets, constants)

    def _filter_to_pipeline(self, filter: MipGenerationFilter) -> ComputePipeline:
        if filter == MipGenerationFilter.AVERAGE:
            return self.avg_pipeline
        elif filter == MipGenerationFilter.AVERAGE_SRGB:
            return self.avg_srgb_pipeline
        elif filter == MipGenerationFilter.MIN:
            return self.min_pipeline
        elif filter == MipGenerationFilter.MAX:
            return self.max_pipeline
        else:
            raise RuntimeError(f"Unhandled mipmap generation filter: {filter}")

    def run(
        self,
        cmd: CommandBuffer,
        image: Image,
        new_layout: ImageLayout,
        view_level_0: Union[Image, ImageView],
        mip_views: List[ImageView],
        instance: SPDPipelineInstance,
        filter: MipGenerationFilter,
    ) -> None:
        s = instance.descriptor_sets.get_current_and_advance()

        s.write_image(view_level_0, ImageLayout.GENERAL, DescriptorType.SAMPLED_IMAGE, 2)
        for m in range(SPD_MAX_LEVELS + 1):
            view = mip_views[m] if m < len(mip_views) else mip_views[0]
            if m == 6:
                s.write_image(view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 3, 0)
            s.write_image(view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 4, m)

        pipeline = self._filter_to_pipeline(filter)
        cmd.bind_compute_pipeline(pipeline, descriptor_sets=[s], push_constants=instance.constants.tobytes())

        cmd.image_barrier(image, ImageLayout.GENERAL, MemoryUsage.ALL, MemoryUsage.COMPUTE_SHADER)
        cmd.dispatch(instance.groups_x, instance.groups_y, image.depth)
        cmd.image_barrier(image, new_layout, MemoryUsage.COMPUTE_SHADER, MemoryUsage.ALL)

    def run_sync_with_views(
        self,
        r: "Renderer",
        image: Image,
        new_layout: ImageLayout,
        view_level_0: Union[Image, ImageView],
        mip_views: List[ImageView],
        filter: MipGenerationFilter,
    ) -> None:
        self.sync_instance.set_image_extents(image.width, image.height, image.mip_levels)
        with r.device.sync_commands() as cmd:
            self.run(cmd, image, new_layout, view_level_0, mip_views, self.sync_instance, filter)

    def run_sync(self, r: "Renderer", image: Image, new_layout: ImageLayout, filter: MipGenerationFilter) -> None:
        level_0_view = ImageView(
            r.device,
            image,
            ImageViewType.TYPE_2D_ARRAY,
            format=Format.R8G8B8A8_SRGB if filter == MipGenerationFilter.AVERAGE_SRGB else Format.R8G8B8A8_UNORM,
            usage_flags=ImageUsageFlags.SAMPLED,
        )
        views = [
            ImageView(
                r.device,
                image,
                ImageViewType.TYPE_2D_ARRAY,
                base_mip_level=m,
                mip_level_count=1,
            )
            for m in range(min(image.mip_levels, SPD_MAX_LEVELS + 1))
        ]
        self.run_sync_with_views(r, image, new_layout, level_0_view, views, filter)

    def run_batched(
        self, cmd: CommandBuffer, mip_generation_requests: Dict[MipGenerationFilter, List[MipGenerationRequest]]
    ) -> List[ImageBarrier]:
        barriers = []
        for filter, requests in mip_generation_requests.items():
            pipeline = self._filter_to_pipeline(filter)
            cmd.bind_pipeline(pipeline)
            for req in requests:
                cmd.bind_descriptor_sets(pipeline, descriptor_sets=[req.descriptor_set])
                cmd.push_constants(pipeline, req.push_constants)
                cmd.dispatch(req.groups_x, req.groups_y, req.groups_z)
                barriers.append(req.after_barrier)
        return barriers



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Tuple

from pyglm.glm import cot, mat3_cast, mat4, orthoRH_ZO, perspectiveRH_ZO, radians, row, vec2, vec3

from .config import Handedness
from .transform3d import RigidTransform3D

# Users:
#   - Application code wants minimal effor to describe camera transformations
#       - Camera controls (FPV / Trackball / orbit)
#       - Camera definition helpers (lookat, ortho, perspective)
#   - Viewer just cares about view and projection transforms -> does not want to recompute or check if changed though
#      -> setter issues where we want to know if parameters change, but do not want to recompute after every change
#         -> V / P can be recomputed once per frame, seems reasonable


@dataclass
class CameraDepth:
    z_near: float
    z_far: float


@dataclass
class Camera:
    camera_from_world: RigidTransform3D
    depth: CameraDepth
    """Horizontal over vertical aspect ratio"""

    def view(self) -> mat4:
        return self.camera_from_world.as_mat4()

    def projection(self, ar: float) -> mat4:
        return mat4(1.0)

    def position(self) -> vec3:
        return self.camera_from_world.inverse().translation

    def right(self) -> vec3:
        r = mat3_cast(self.camera_from_world.rotation)
        return vec3(-row(r, 0))

    def up(self) -> vec3:
        r = mat3_cast(self.camera_from_world.rotation)
        return vec3(row(r, 1))

    def front(self) -> vec3:
        r = mat3_cast(self.camera_from_world.rotation)
        return vec3(-row(r, 2))

    def right_up_front(self) -> Tuple[vec3, vec3, vec3]:
        r = mat3_cast(self.camera_from_world.rotation)
        return vec3(-row(r, 0)), vec3(row(r, 1)), vec3(-row(r, 2))

    def film_dist(self) -> float:
        return 0.0


@dataclass
class OrthographicCamera(Camera):
    center: vec2
    half_extents: vec2

    @classmethod
    def look_at(
        cls,
        position: vec3,
        target: vec3,
        up: vec3,
        z_near: float,
        z_far: float,
        center: vec2,
        half_extents: vec2,
        handedness: Handedness,
    ) -> "OrthographicCamera":
        return cls(
            camera_from_world=RigidTransform3D.look_at(position, target, up, handedness),
            depth=CameraDepth(z_near, z_far),
            center=center,
            half_extents=half_extents,
        )

    def projection(self, ar: float) -> mat4:
        half_extents = self.half_extents * vec2(ar, 1)
        top_left = self.center - half_extents
        bottom_right = self.center + half_extents
        return orthoRH_ZO(
            top_left.x,
            bottom_right.x,
            bottom_right.y,
            top_left.y,
            self.depth.z_near,
            self.depth.z_far,
        )


@dataclass
class PerspectiveCamera(Camera):
    fov: float
    """Vertical field of view in radians"""

    @classmethod
    def look_at(
        cls,
        position: vec3,
        target: vec3,
        up: vec3,
        z_near: float,
        z_far: float,
        fov: float,
        handedness: Handedness,
    ) -> "PerspectiveCamera":
        return cls(
            camera_from_world=RigidTransform3D.look_at(position, target, up, handedness),
            depth=CameraDepth(z_near, z_far),
            fov=fov,
        )

    def projection(self, ar: float) -> mat4:
        return perspectiveRH_ZO(
            radians(self.fov),
            ar,
            self.depth.z_near,
            self.depth.z_far,
        )

    def film_dist(self) -> float:
        return 0.5 * cot(radians(self.fov) * 0.5)



import av
import numpy as np
from numpy.typing import NDArray

from ambra.config import Config, GuiConfig
from ambra.primitives3d import Lines
from ambra.viewer import Viewer

width, height = 1920, 1080
viewer = Viewer(
    config=Config(
        window=False,
        window_width=width,
        window_height=height,
        gui=GuiConfig(stats=True),
    ),
)

positions = np.array(
    [
        [0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0],
        [0.0, 0.0, 1.0],
    ],
    np.float32,
)

colors = np.array(
    [
        0xFF0000FF,
        0xFF0000FF,
        0xFF00FF00,
        0xFF00FF00,
        0xFFFF0000,
        0xFFFF0000,
    ],
    np.uint32,
)

line = Lines(positions, colors, 1.0, translation=np.linspace((0, 0, 0), (1, 0, 0), 50))
viewer.scene.objects.extend([line])


container = av.open("test.mp4", mode="w")

stream = container.add_stream("h264", rate=viewer.playback.frames_per_second, options={})
stream.width = width
stream.height = height
stream.pix_fmt = "yuv420p"


def on_frame(img: NDArray[np.uint8]):
    frame = av.VideoFrame.from_ndarray(img[:, :, :3], format="rgb24")
    for packet in stream.encode(frame):
        container.mux(packet)


viewer.render_video(on_frame)

# Flush stream
for packet in stream.encode():
    container.mux(packet)

# Close the file
container.close()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from typing import List, Optional, Tuple

from pyxpg import (
    DescriptorPool,
    DescriptorPoolCreateFlags,
    DescriptorPoolSize,
    DescriptorSet,
    DescriptorSetBinding,
    DescriptorSetLayout,
    DescriptorSetLayoutCreateFlags,
    Device,
)

from .ring_buffer import RingBuffer


def create_descriptor_layout_pool_and_set(
    device: Device,
    bindings: List[DescriptorSetBinding],
    variable_count: int = 0,
    layout_flags: DescriptorSetLayoutCreateFlags = DescriptorSetLayoutCreateFlags.NONE,
    pool_flags: DescriptorPoolCreateFlags = DescriptorPoolCreateFlags.NONE,
    name: Optional[str] = None,
) -> Tuple[DescriptorSetLayout, DescriptorPool, DescriptorSet]:
    layout = DescriptorSetLayout(device, bindings, layout_flags, name=name)
    pool = DescriptorPool(device, [DescriptorPoolSize(b.count, b.type) for b in bindings], 1, pool_flags, name=name)
    set = pool.allocate_descriptor_set(layout, variable_count, name=name)
    return layout, pool, set


def create_descriptor_layout_pool_and_sets(
    device: Device,
    bindings: List[DescriptorSetBinding],
    count: int,
    variable_count: int = 0,
    layout_flags: DescriptorSetLayoutCreateFlags = DescriptorSetLayoutCreateFlags.NONE,
    pool_flags: DescriptorPoolCreateFlags = DescriptorPoolCreateFlags.NONE,
    name: Optional[str] = None,
) -> Tuple[DescriptorSetLayout, DescriptorPool, List[DescriptorSet]]:
    layout = DescriptorSetLayout(device, bindings, layout_flags, name=name)
    pool = DescriptorPool(
        device, [DescriptorPoolSize(b.count * count, b.type) for b in bindings], count, pool_flags, name=name
    )
    sets: List[DescriptorSet] = [pool.allocate_descriptor_set(layout, variable_count, name=name) for _ in range(count)]
    return layout, pool, sets


def create_descriptor_layout_pool_and_sets_ringbuffer(
    device: Device,
    bindings: List[DescriptorSetBinding],
    count: int,
    variable_count: int = 0,
    layout_flags: DescriptorSetLayoutCreateFlags = DescriptorSetLayoutCreateFlags.NONE,
    pool_flags: DescriptorPoolCreateFlags = DescriptorPoolCreateFlags.NONE,
    name: Optional[str] = None,
) -> Tuple[DescriptorSetLayout, DescriptorPool, RingBuffer[DescriptorSet]]:
    layout, pool, sets = create_descriptor_layout_pool_and_sets(
        device, bindings, count, variable_count, layout_flags, pool_flags, name
    )
    return layout, pool, RingBuffer(sets)


def create_descriptor_pool_and_sets(
    device: Device,
    layout: DescriptorSetLayout,
    count: int,
    variable_count: int = 0,
    pool_flags: DescriptorPoolCreateFlags = DescriptorPoolCreateFlags.NONE,
    name: Optional[str] = None,
) -> Tuple[DescriptorPool, List[DescriptorSet]]:
    pool = DescriptorPool(
        device, [DescriptorPoolSize(b.count * count, b.type) for b in layout.bindings], count, pool_flags, name=name
    )
    sets: List[DescriptorSet] = [pool.allocate_descriptor_set(layout, variable_count, name=name) for _ in range(count)]
    return pool, sets


def create_descriptor_pool_and_sets_ringbuffer(
    device: Device,
    layout: DescriptorSetLayout,
    count: int,
    variable_count: int = 0,
    pool_flags: DescriptorPoolCreateFlags = DescriptorPoolCreateFlags.NONE,
    name: Optional[str] = None,
) -> Tuple[DescriptorPool, RingBuffer[DescriptorSet]]:
    pool, sets = create_descriptor_pool_and_sets(device, layout, count, variable_count, pool_flags, name)
    return pool, RingBuffer(sets)



from pyxpg import Action, Key, Modifiers, imgui

from ambra.config import Config, ServerConfig
from ambra.server import Client, MessageId, RawMessage
from ambra.utils.hook import hook
from ambra.viewer import Viewer


class Viewer(Viewer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.messages = []

    @hook
    def on_gui(self):
        if imgui.begin("Window")[0]:
            for message in self.messages:
                imgui.text(message)
        imgui.end()

    @hook
    def on_key(self, key: Key, action: Action, modifiers: Modifiers):
        pass

    def on_raw_message(self, client: Client, raw_message: RawMessage):
        print(
            f"Message received from: {client.name} ({client.address}, {client.port}): {raw_message.id} {raw_message.format} {MessageId.USER.value}"
        )
        if raw_message.id == MessageId.USER.value:
            print("Decoding..")
            self.messages.append(raw_message.data.decode("utf-8"))
        else:
            super().on_raw_message(client, raw_message)


config = Config(
    server=ServerConfig(enabled=True),
)
viewer = Viewer("Server", config=config)
viewer.run()
